from .media import files_plot, media_plot, chat_media_plot, chat_files_plot
from .share import share_plot, chat_share_plot
from .emoji import chat_emoji, all_emoji, emojis, Multi_emojis
from .words import Words, Multi_Words

from .utils import image_formatter, get_thumbnail
from .plot_helper import get_meta_data, metric_correct, cat_filter
from .common_plots import plot, chat_plot
import pandas as pd
from .get_df import df_from_jsons, set_dur_unit
import tldextract
import os
import shutil
import platform

from tqdm import tqdm
import json
import emoji
import re
import numpy as np
from pandas.util._decorators import doc

cwd = os.path.dirname(__file__)
metric_to_cat = {"share": "domain", "files": "ext", "media_type": "media"}


def set_meta_data(meta_data, cwd, filename="meta_data.json"):
    json_object = json.dumps(meta_data, indent=4)
    with open(os.path.join(cwd, filename), "w") as outfile:
        outfile.write(json_object)


@doc()
def describe(self, n=5, others=False):
    """
    Similar to to pandas.DataFrame.describe

    Generate descriptive statistics of the data for every chat included 
    and for every sender in that chat

    Parameters
    ----------
    n : int , default 5
        number of chats to included,
        it will filter the data by default for only the Top n chats
    others: bool, default False
        whether to keep the chats other than the Top n or not
        When True, the other chats are grouped together and labeled
        as Others

    Returns
    -------
    Pandas.DataFrame

    """
    meta_data = get_meta_data()
    df = self.data.copy()
    do_sum = self.plot._metric in ["call_duration", "audio_files", "video"]
    metric = metric_correct(self.plot._metric, do_sum and "sum" or "count")
    if metric in metric_to_cat.keys():
        metric = metric_to_cat[metric]
    if df.chat.nunique() > 1:
        df, _ = cat_filter(df, n, metric=metric, others=others)

    if others:

        df.loc[
            (df.chat == "Others") & (df.sender_name != meta_data["user"]), "sender_name"
        ] = "Others"
    order = {key: value for (value, key) in enumerate(df.chat.value_counts().index)}
    data = df.groupby(["chat", "sender_name"])[metric].describe()
    if do_sum:
        data["sum"] = df.groupby(["chat", "sender_name"])[metric].sum()
    for chat in data.index.levels[0]:
        chat_data = df[df.chat == chat]
        overall_row = tuple(chat_data[metric].describe())
        if do_sum:
            overall_row = overall_row + (chat_data[metric].sum(),)
        data.loc[(chat, "#Overall"), :] = overall_row
    data = data.sort_index(level=0, key=lambda x: x.map(order))

    if df.chat.nunique() > 1:

        overall_row = tuple(df[metric].describe())
        if do_sum:
            overall_row = overall_row + (df[metric].sum(),)
        data.loc[("#Overall", "#Overall"), :] = overall_row

        df["sender_name"] = df["sender_name"].map(
            lambda name: name if name in [meta_data["user"]] else "Others"
        )
        for sender, row in df.groupby("sender_name")[metric].describe().iterrows():
            overall_row = tuple(row)
            if do_sum:
                overall_row = overall_row + (df[metric].sum(),)
            data.loc[("#Overall", sender), :] = overall_row

    if not do_sum:
        if metric not in ["media", "sticker", "ext", "domain"]:
            data.drop(columns=["unique", "top", "freq"], inplace=True)
        elif metric == "sticker":
            base = get_meta_data()["data_dir"]
            try:
                get_ipython().__class__.__name__
                from IPython.display import HTML
            except:
                return data
            try:
                data["top"] = data["top"].map(lambda uri: os.path.join(base, uri))
                data["top"] = data.top.map(get_thumbnail)
            except:
                return data

            return HTML(data.to_html(formatters={"top": image_formatter}, escape=False))
    else:
        data = data.apply(round, args=[2])
    return data

def detect_languages(data):
    if not data.content.count():
        raise ValueError('No text data exist')
    if platform.system() == "Linux":
        try:
            from polyglot.detect import Detector
        except ModuleNotFoundError:
            raise ModuleNotFoundError("""polyglot is required for language detection,install it using `pip install polyglot`.
                please note that polyglot requires  PyICU and pycld2 in order to operate correctly, consider installing them
                using `pip install PyICU`, `pip install pycld2`""")

        from operator import attrgetter
        get_name=attrgetter('name')
        get_per=attrgetter('confidence')
        languages = [
                {get_name(lan): f"{get_per(lan)}%"}
                for lan in Detector("".join(data.content_striped.dropna())).languages
                if get_per(lan)]
    else:

        try :
            import cld2
        except ModuleNotFoundError:
            raise ModuleNotFoundError("""cld2 is required for language detection,install it using `pip install cld2-cffi`.""")

        languages = [
                {lan[0]: f"{lan[2]}%"}
                for lan in cld2.detect("".join(data.content_striped.dropna()))[2]
                if lan[2]]
    return languages

plot_type = {"single": chat_plot, "multi": plot}
files_pt = {"single": chat_files_plot, "multi": files_plot}
media_pt = {"single": chat_media_plot, "multi": media_plot}


@doc(klass="Photos", other_attrs="", other_methods="")
class Photos:
    """
    The Class that enables you to deal with {klass} data

    Attributes
    ----------
    data : pandas.DataFrame
        The {klass} row data
    count : int
        The total number of {klass} you have sent or recieved
    plot : obj
        The obj that you use to plot your data
    {other_attrs}

    Methods
    -------
    describe(n=5,others=False)
        Generate descriptive statistics
    {other_methods}
    """

    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.data["photos"] = self.data["photos"].apply(
            lambda item: item["uri"] if item == item else np.nan
        )
        self.count = self.data.photos.count()
        self.plot = plot_type[p_type]("photos", self.data)

    @doc(describe)
    def describe(self, n=5, others=False):

        return describe(self, n, others)


@doc(Photos, klass="Gifs", other_attrs="", other_methods="")
class Gifs:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.data["gifs"] = self.data["gifs"].apply(
            lambda item: item["uri"] if item == item else np.nan
        )
        self.count = self.data.gifs.count()
        self.plot = plot_type[p_type]("gifs", self.data)

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


@doc(
    Photos,
    klass="Files",
    other_attrs="exts : pandas.Series \n    the file extensions you used and how often",
    other_methods="",
)
class Files:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.data["files"] = self.data["files"].apply(
            lambda item: item["uri"] if item == item else np.nan
        )
        self.data["ext"] = self.data["files"].apply(
            lambda file: os.path.splitext(file)[1]
        )
        self.count = self.data.files.count()
        self.exts = self.data.ext.value_counts()
        self.plot = files_pt[p_type]("files", self.data)

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


@doc(Photos, klass="Stickers", other_attrs="", other_methods="")
class Stickers:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.data["sticker"] = self.data["sticker"].apply(
            lambda item: item["uri"] if item == item else np.nan
        )
        self.count = self.data.sticker.count()
        self.plot = plot_type[p_type]("sticker", self.data)

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


@doc(
    Photos,
    klass="Audio Files",
    other_attrs="""total_audio_files_dur : int
    The total length of all of your audio files
longest_audio_file : pandas.Series
    Information about the audio file with the longest duration
curr_time_unit : str
    the current time unit of your audio files length""",
    other_methods=""""change_dur_unit(unit)
    Change the time unit of audio files length""",
)
class Audios:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.data["audio_files_uri"] = self.data.audio_files.apply(
            lambda item: item["uri"] if item == item else np.nan
        )
        self.data["audio_files_length"] = self.data.audio_files.apply(
            lambda item: item["length"] if item == item else np.nan
        )
        self.data.drop(columns=["audio_files"], inplace=True)
        self.count = self.data.audio_files_uri.count()
        self.plot = plot_type[p_type]("audio_files", self.data)
        self.__duration_unit = get_meta_data()["time_unit"]

    @property
    def total_audio_files_dur(self):
        return self.data.audio_files_length.sum()

    @property
    def longest_audio_file(self):
        return self.data.loc[self.data.audio_files_length.idxmax()]

    @property
    def curr_time_unit(self):
        return self.__duration_unit

    @doc(metric="audio")
    def change_dur_unit(self, unit):
        """
        Used to change the time unit of your {metric} data duration

        Parameters
        ----------
        unit : ('sec','minute','hour','day')
            the time unit to convert the data into
        """
        if unit == self.__duration_unit:
            return
        self.data.audio_files_length = set_dur_unit(
            self.data.audio_files_length, self.__duration_unit, unit
        )
        self.__duration_unit = unit

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


@doc(
    Photos,
    klass="Videos",
    other_attrs="""total_video_dur : int
    The total length of all of your videos
longest_video : pandas.Series
    Information about the video with the longest duration
curr_time_unit : str
    the current time unit of your videos length """,
    other_methods="change_dur_unit(unit)\n    Change the time unit of videos length",
)
class Videos:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.data["video_uri"] = self.data.videos.apply(
            lambda item: item["uri"] if item == item else np.nan
        )
        self.data["video_length"] = self.data.videos.apply(
            lambda item: item["duration"] if item == item else np.nan
        )
        # self.data["video_frame_count"]=self.data.videos.apply(lambda item: item["frame_count"] if item==item else np.nan)
        self.data.drop(columns=["videos"], inplace=True)
        self.count = self.data.video_uri.count()
        self.plot = plot_type[p_type]("video", self.data)
        self.__duration_unit = get_meta_data()["time_unit"]

    @property
    def curr_time_unit(self):
        return self.__duration_unit

    @property
    def total_video_dur(self):
        return self.data.video_length.sum()

    @property
    def longest_video(self):
        return self.data.loc[self.data.video_length.idxmax()]

    @doc(Audios.change_dur_unit, metric="video")
    def change_dur_unit(self, unit):
        if unit == self.__duration_unit:
            return
        self.data.video_length = set_dur_unit(
            self.data.video_length, self.__duration_unit, unit
        )
        self.__duration_unit = unit

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


@doc(
    Photos,
    klass="The Whole Media",
    other_attrs="used_media : pandas.Series \n    the types of media you used and how often",
    other_methods="",
)
class Media:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        media_files = {}
        for media in ["audio_files", "videos", "photos", "gifs", "files"]:
            media_files[media] = (
                self.data[["timestamp_ms", "chat", "sender_name", media]]
                .dropna()
                .explode(column=media)
            )

        media_files["sticker"] = self.data[
            ["timestamp_ms", "chat", "sender_name", "sticker"]
        ].dropna()

        if len(media_files["audio_files"]):
            self.Audios = Audios(media_files["audio_files"].copy(), p_type)
        if len(media_files["videos"]):
            self.Videos = Videos(media_files["videos"].copy(), p_type)
        if len(media_files["photos"]):
            self.Photos = Photos(media_files["photos"].copy(), p_type)
        if len(media_files["gifs"]):
            self.Gifs = Gifs(media_files["gifs"].copy(), p_type)
        if len(media_files["files"]):
            self.Files = Files(media_files["files"].copy(), p_type)
        if len(media_files["sticker"]):
            self.Stickers = Stickers(media_files["sticker"].copy(), p_type)

        for media in media_files.keys():
            media_files[media][media] = media

        self.data = (
            pd.concat(media_files.values())
            .melt(
                id_vars=["timestamp_ms", "chat", "sender_name"],
                value_vars=media_files.keys(),
                var_name="media_type",
            )
            .dropna()
        )
        self.data["value"] = self.data.media_type.replace(
            {
                "audio_files": "Audio",
                "videos": "Video",
                "photos": "Photo",
                "gifs": "GIF",
                "files": "File",
                "sticker": "Sticker",
            }
        )
        self.count = self.data.media_type.count()
        self.used_media = self.data.media_type.value_counts()
        self.data.rename(columns={"value": "media"}, inplace=True)
        if p_type == "single":
            self.data.drop(columns=["chat"])
        self.plot = media_pt[p_type]("media_type", self.data)

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


emoji_pt = {"single": emojis, "multi": Multi_emojis}
words_pt = {"single": Words, "multi": Multi_Words}
content_pt = {"single": chat_plot, "multi": plot}


@doc(
    Photos,
    klass="Messages content",
    other_attrs="""Emoji : obj
    The obj that enables you to deal with the emojis in your messages
Words : obj
    The obj that enables you to deal with the text of your messages""",
    other_methods="",
)
class Content:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.count = self.data.content.count()
        if self.data.emoji.count():
            self.Emoji = emoji_pt[p_type](passed_data)
        self.Words = words_pt[p_type](
            passed_data[
                ["timestamp_ms", "chat", "sender_name", "content", "content_striped"]
            ]
        )
        self.plot = content_pt[p_type](
            "content", passed_data[["timestamp_ms", "chat", "sender_name", "content"]]
        )

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


calls_pt = {"single": chat_plot, "multi": plot}


@doc(
    Photos,
    klass="Calls",
    other_attrs="""first_call_date : timestamp 
    The date in which the first call were made
last_call_date : timestamp
    The date in which the first call were made
total_calls_dur : float
    The total duration of all calls
total_video_dur : float
    The total duration of all video calls
total_voice_dur : float
    The total duration of all voice calls
longest_call : pandas.Series
    Information about the call with the longest duration
curr_time_unit : str
    the current time unit of your call duration """,
    other_methods="""change_dur_unit(unit)
    Change the time unit of call duration""",
)
class Calls:
    def __init__(self, passed_data, p_type):
        self.data = passed_data
        self.count = self.data.call_duration.count()
        self.first_call_date = self.data.timestamp_ms.min()
        self.last_call_date = self.data.timestamp_ms.max()

        if p_type == "multi":
            self.most_called = (
                self.data.groupby("chat")["call_duration"].count().idxmax()
            )
        # or self.data.chat.value_counts()[0].index
        self.plot = calls_pt[p_type](
            "call_duration",
            passed_data[
                [
                    "timestamp_ms",
                    "chat",
                    "sender_name",
                    "call_duration",
                    "call_status",
                    "call_type",
                ]
            ],
        )
        self.__duration_unit = get_meta_data()["time_unit"]

    @property
    def curr_time_unit(self):
        return self.__duration_unit

    @property
    def total_calls_dur(self):
        return self.data.call_duration.sum()

    @property
    def total_video_dur(self):
        return self.data[self.data.call_type == "video"].call_duration.sum()

    @property
    def total_voice_dur(self):
        return self.data[self.data.call_type == "voice"].call_duration.sum()

    @property
    def longest_call(self):
        return self.data.loc[self.data.call_duration.idxmax()]

    @doc(Audios.change_dur_unit, metric="call")
    def change_dur_unit(self, unit):
        if unit == self.__duration_unit:
            return
        self.data.call_duration = set_dur_unit(
            self.data.call_duration, self.__duration_unit, unit
        )
        self.__duration_unit = unit

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


links_pt = {"single": chat_share_plot, "multi": share_plot}


@doc(
    Photos,
    klass="Share Links",
    other_attrs="all_domains : pandas.Series \n    the domains of links you shared and their counts",
    other_methods="",
)
class links:
    def __init__(self, passed_data, p_type):
        self.data = passed_data.copy()
        self.data["share"] = self.data["share"].apply(
            lambda item: item["link"] if item == item else np.nan
        )
        self.data["domain"] = (
            self.data["share"].apply(tldextract.extract).apply(lambda row: row.domain)
        )
        self.count = self.data.share.count()
        self.all_domains = self.data.domain.value_counts()
        self.plot = links_pt[p_type]("share", self.data)

    @doc(describe)
    def describe(self, n=5, others=False):
        return describe(self, n, others)


class Chat:
    """
    The class for single chat
    Use [chat_name] to access a chat instance. 

    Attributes
    ----------
    data : pandas.DataFrame
        A subset of all data for a single chat
    name : str
        Chat name as it appears in messenger
    n_of_messages : int
        The count of all messages in this chat
    languages : dict
        The languages used in this chat and their percentages
    chat_starter : str
        The name of the person who started the conversation
    first_msg_date : timestamp
        The date of the first message
    n_of_distinct_days : int
        The total number of different days you chatted on
    longest_streak : int
        Longest number of days of continuous chatting/chatting streak
    streak_periods : dict
        The period in which you achieved the longest streak
    last_hear : timestamp
        The time of the last message between you
    longest_msgs : pandas.DataFrame
        Information about the longest 10 messages in terms of word count
    Media : Media obj
        to deal with chat Media, only exist if the chat have at least one media file
    Calls : Calls obj
        to deal with chat Calls, only exist if you did at least one call 
    Content : Content obj
        to deal with chat messages content, only exist if you have at least one message between you
    Reacts : Reacts obj
        to deal with chat Reacts, only exist if at least one of you reacted with a message
    Links : Links obj
        to deal with Links you shared in the chat, only exist if you have sent at least one link between you
    plot : Plot obj
        used to plot your data

    """

    def __init__(self, passed_name, passed_data, SubModules):
        if hasattr(self, "data"):
            self.__reload(SubModules)
        else:
            self.data = passed_data
            self.name = passed_name
            self.n_of_messages = len(self.data)
            if self.data.content.count():
                self.longest_msgs = (
                    self.data.join(
                        self.data.content.dropna().apply(len), rsuffix="_len"
                    )
                    .sort_values("content_len", ascending=False)[
                        ["timestamp_ms", "sender_name", "content", "content_len"]
                    ][:10]
                    .reset_index(drop=True)
                )
            # self.reactions=self.data['reactions'].dropna().str.split().explode().value_counts()
            self.chat_starter = (
                self.data.sort_values("timestamp_ms").iloc[0].sender_name
            )
            self.first_msg_date = str(self.data.timestamp_ms.min())
            self.n_of_distinct_days = self.data.timestamp_ms.dt.date.nunique()
            unique_days = pd.Series(
                sorted(self.data.timestamp_ms.dt.date.unique()), name="dates"
            )
            if len(unique_days) <= 1:
                self.longest_streak = 1
                self.streak_periods = {
                    "Start": unique_days.min().strftime("%Y-%m-%d"),
                    "End": unique_days.min().strftime("%Y-%m-%d"),
                }
            else:
                consec_dates = unique_days.to_frame("dates").join(
                    (
                        unique_days.diff()
                        .dt.days.ne(1)
                        .cumsum()
                        .to_frame()
                        .groupby("dates")
                        .cumcount()
                        + 1
                    ).rename("n_consec")
                )
                self.longest_streak = consec_dates["n_consec"].max()
                periods = []
                for date in consec_dates[
                    consec_dates.n_consec == self.longest_streak
                ].dates:
                    periods.append(
                        {
                            "Start": (
                                date - pd.Timedelta(days=self.longest_streak - 1)
                            ).strftime("%Y-%m-%d"),
                            "End": date.strftime("%Y-%m-%d"),
                        }
                    )
                self.streak_periods = periods
            # self.longest_streak=len(unique_days)>1 and (pd.Series(sorted(unique_days)).diff().dt.days.ne(1).cumsum().to_frame().groupby(0).cumcount()+1).max() or 1
            self.last_hear = str(self.data.timestamp_ms.max())
            self.plot = chat_plot("chat", self.data)
            self.__reload(SubModules)
    @property
    def languages(self):
        return detect_languages(self.data)
    @doc(describe)
    def describe(self, n=5, others=False):

        return describe(self, n, others)
    def __reload(self, SubModules):

        if (
            self.data[["photos", "sticker", "files", "videos", "audio_files", "gifs"]]
            .count()
            .max()
            > 0
            and (SubModules == "All" or "Media" in SubModules)
            and not hasattr(self, "Media")
        ):
            self.Media = Media(
                self.data[
                    [
                        "timestamp_ms",
                        "chat",
                        "sender_name",
                        "photos",
                        "sticker",
                        "files",
                        "videos",
                        "audio_files",
                        "gifs",
                    ]
                ],
                "single",
            )
        # self.total_voice_dur=hasattr(self,'Media') and hasattr(self.Media,'Audios') and self.Media.Audios.data.audio_files_length.sum()//60 or 0
        if (
            self.data.call_duration.count()
            and (SubModules == "All" or "Calls" in SubModules)
            and not hasattr(self, "Calls")
        ):
            self.Calls = Calls(
                self.data[
                    [
                        "timestamp_ms",
                        "chat",
                        "sender_name",
                        "call_duration",
                        "call_status",
                        "call_type",
                    ]
                ].dropna(),
                "single",
            )
        if (
            self.data.content.count()
            and (SubModules == "All" or "Content" in SubModules)
            and not hasattr(self, "Content")
        ):
            self.Content = Content(
                self.data[
                    [
                        "timestamp_ms",
                        "chat",
                        "sender_name",
                        "content",
                        "content_striped",
                        "emoji",
                    ]
                ].dropna(subset="content"),
                "single",
            )
        if (
            self.data.reactions.count()
            and (SubModules == "All" or "Reacts" in SubModules)
            and not hasattr(self, "Reacts")
        ):
            self.Reacts = chat_emoji(
                "reactions",
                self.data[
                    ["timestamp_ms", "chat", "sender_name", "reactions"]
                ].dropna(),
            )
            # self.Reacts=Reacts(,'single')
        if (
            self.data.share.count()
            and (SubModules == "All" or "Links" in SubModules)
            and not hasattr(self, "Links")
        ):
            self.Links = links(
                self.data[["timestamp_ms", "chat", "sender_name", "share"]].dropna(),
                "single",
            )
            

class Stats:
    """
    The class for single chat

    Attributes
    ----------
    data : pandas.DataFrame
        the full data
    n_of_messages : int
        The count of all messages 
    top_chats : list
        The top 5 chats in terms of number of messages
    languages : dict
        The all languages used in your chat and their percentages -- limted to only 3 languages
    first_msg_date : timestamp
        The date of the first message
    first_chat : str 
        The first ever person you chatted with
    top_streaks : pandas.Series
        Top 10 Longest chatting streak, one for each chat
    longest_msgs : pandas.DataFrame
        Information about the longest 10 messages in terms of word count
    Media : Media obj
        to deal with Media, only exist if the data contains at least one media file
    Calls : Calls obj
        to deal with Calls, only exist if you did/recieved at least one call 
    Content : Content obj
        to deal with messages content, only exist if your data have at least one message
    Reacts : Reacts obj
        to deal with Reacts, only exist if at least one react is made
    Links : Links obj
        to deal with shared Links, only exist if at least one link is shared
    plot : Plot obj
        used to plot your data

    Methods
    -------
    __init__(base='./',restore=False,**kwargs)
        Creates the data
    load(SubModules=False,chat_instances=True)
        Build the main classes and create chat instances
    freeze()
        to save your data to be restored in a later time
    update(base='./',after=True,**kwargs)
        to extend your data


    """

    def __init__(self, base="./", restore=False, **kwargs):
        """
        Generates a DataFrame from your row data or restore the df you saved

        Parameters
        ----------
        base : str, default cwd
            The the directory where the data is located.
            Only passed when generating the data for the first time,
            has no effect when `restore` is True.
        restore : bool, default False
            if you saved the data using `freeze()` method, you can restore it without the need to regenerating it.
        **kwargs : parameters to control data generation
            parallel : bool , default False
                When True , data is processed in parallel whenever it possible.
                It helps speed up the process of generating the data, but it consumes your computer resources.
            max_workers : int, default 2
                The number of threads to use when processing the data in parallel.
            process_audio : bool, default True
                whether to process the audio files to get their durtion or not.
                Processing audio files is a computationally intensive process, 
                especially if your data contains a lot of audio files,
                you can speed up this process by passing {parallel = True} to process audio data in parallel.
            dur_unit : {'sec','minute','hour','day'}, default 'sec'
                The time unit of the length of calls, videos, and audio files
        """
        self.__Chats = {}
        # self.Chats=locals()
        if restore:
            if restore == "sample":
                meta_data = get_meta_data("sample_meta_data.json")
                meta_data["data_dir"] = cwd
                set_meta_data(meta_data, cwd, "sample_meta_data.json")
            data_name = restore == "sample" and "SAMPLE_DATA.pkl" or "MY_DATA.pkl"
            if os.path.isfile(os.path.join(cwd, data_name)):
                self.data = pd.read_pickle(os.path.join(cwd, data_name))
                shutil.copy(
                    os.path.join(
                        cwd,
                        restore == "sample"
                        and "sample_meta_data.json"
                        or "gen_meta_data.json",
                    ),
                    os.path.join(cwd, "meta_data.json"),
                )
                meta_data = get_meta_data()
                self.user = meta_data["user"]
                # os.path.splitext(file)[1]
            else:
                raise FileNotFoundError(f"No data to restore")

        else:
            if not os.path.exists(base):
                raise FileNotFoundError("directory not found")
            dur_unit = kwargs.get("dur_unit", "sec")
            self.data = df_from_jsons(base, **kwargs)

            info = json.load(
                open(os.path.join(base, "messages\\autofill_information.json"), "r")
            )
            self.user = info[list(info)[-1]]["FULL_NAME"][0]
            meta_data = {
                "user": self.user,
                "Top": [],
                "data_dir": base,
                "time_unit": dur_unit,
            }
            set_meta_data(meta_data, cwd, "gen_meta_data.json")
        self.__loaded = False

    @property
    def first_msg_date(self):
        return str(self.data.timestamp_ms.min())

    @property
    def last_msg_date(self):
        return str(self.data.timestamp_ms.max())

    def __getitem__(self, name):
        if not len(self.__Chats):
            raise Exception("Please load chat_instances first")
        if name not in self.__Chats.keys():
            raise KeyError(f"chat : {name}, not found in your data")
        return self.__Chats[name]
        # return self.Chats[arg]

    # def __getattr__(self,attr):
    # return self.Chats[attr]
    @doc(describe)
    def describe(self, n=5, others=False):

        return describe(self, n, others)
    def load(self, SubModules="All", chat_instances=True):
        """
        Build the main classes and create the chat instances.
        Without load, you wont be able to use or access any of Stats attributes/Modules except `df` attribue.

        Parameters
        ----------
        SubModules : {'Media','Calls','Content','Reacts','Links'} or a list of more than one them, default 'All'
            When 'All', All Modules are loaded.
            When [] : No Module will be loaded
            if specified, only the Modules you specified are loaded.
            Can be used to reduce the load time if you don't need to use all Modules.
        chat_instances : bool, default True
            Whether to create chat instances or not
            Can be used to reduce the load time if you are not willing to inspect the data of individual chats.
        """
        if not len(self.data):
            raise Exception("data not exist")
        if SubModules and SubModules != "All":
            if not isinstance(SubModules, list):
                SubModules = [SubModules]
            for SubModule in SubModules:
                if SubModule not in ["Media", "Calls", "Content", "Reacts", "Links"]:
                    raise ValueError(f"No Module called {SubModule},")
        if SubModules:
            print("Building Main Modules ...")
        if self.__loaded:
            self.__reload(SubModules, chat_instances)
        else:
            self.n_of_messages = len(self.data)
            self.top_chats = self.data.chat.value_counts()[:5].index.tolist()
            meta_data = get_meta_data()
            meta_data["Top"] = self.top_chats
            set_meta_data(meta_data, cwd)

            if self.data.content.count():

                self.data.content = (
                    self.data[
                        ~(
                            (self.data.content.dropna().str.contains("تم التفاعل"))
                            | ~self.data.call_duration.isna()
                            | self.data.content.dropna().str.contains(
                                "لقد أرسلتَ مرفقًا."
                            )
                            | self.data.content.dropna().str.contains(
                                "تم إرسال مرفق بواسطة"
                            )
                        )
                    ]
                    .content.reindex(range(len(self.data)))
                    .map(
                        lambda txt: re.sub(r"http\S+", "", txt), na_action='ignore'
                    )
                    .replace("", np.NaN)
                )

                self.data["content_striped"] = (
                    self.data.content.map(
                        lambda text: emoji.replace_emoji(text, ""), na_action='ignore'
                    )
                    .str.strip()
                    .replace("", np.nan)
                )


                self.longest_msgs = (
                    self.data.join(self.data.content.dropna().apply(len), rsuffix="_len")
                    .sort_values("content_len", ascending=False)[
                        [
                            "timestamp_ms",
                            "chat",
                            "sender_name",
                            "content",
                            "content_len",
                        ]
                    ][:10]
                    .reset_index(drop=True)
                )

            self.first_chat = self.data.sort_values("timestamp_ms").iloc[0].chat
            self.top_streaks = (
                self.data.groupby(["chat", pd.Grouper(key="timestamp_ms", freq="D")])
                .size()
                .reset_index()
                .groupby("chat")["timestamp_ms"]
                .unique()
                .map(
                    lambda dates: (
                        pd.Series(sorted(dates), name="dates")
                        .diff()
                        .dt.days.ne(1)
                        .cumsum()
                        .to_frame()
                        .groupby("dates")
                        .cumcount()
                        + 1
                    ).max()
                )
                .sort_values(ascending=False)[:10]
            )
            # self.reactions=self.data['reactions'].dropna().str.split().explode().value_counts()

            self.__reload(SubModules, chat_instances)
            self.plot = plot("index", self.data.reset_index())
            self.__loaded = True
    @property
    def languages(self):
        return detect_languages(self.data)

    def freeze(self):
        """
        Saves your data to be restored in a later time
        """
        self.data.to_pickle(os.path.join(cwd, "MY_DATA.pkl"))

    def __reload(self, SubModules=False, chat_instances=True):

        if (
            self.data[["photos", "sticker", "files", "videos", "audio_files", "gifs"]]
            .count()
            .max()
            > 0
            and (SubModules == "All" or "Media" in SubModules)
            and not hasattr(self, "Media")
        ):
            print("Creating Media ...")
            self.Media = Media(
                self.data[
                    [
                        "timestamp_ms",
                        "chat",
                        "sender_name",
                        "photos",
                        "sticker",
                        "files",
                        "videos",
                        "audio_files",
                        "gifs",
                    ]
                ],
                "multi",
            )
        if (
            self.data.call_duration.count()
            and (SubModules == "All" or "Calls" in SubModules)
            and not hasattr(self, "Calls")
        ):
            print("Creating Calls ...")
            self.Calls = Calls(
                self.data[
                    [
                        "timestamp_ms",
                        "chat",
                        "sender_name",
                        "call_duration",
                        "call_status",
                        "call_type",
                    ]
                ].dropna(),
                "multi",
            )
        if (
            self.data.content.count()
            and (SubModules == "All" or "Content" in SubModules)
            and not hasattr(self, "Content")
        ):
            print("Creating Content ...")
            self.Content = Content(
                self.data[
                    [
                        "timestamp_ms",
                        "chat",
                        "sender_name",
                        "content",
                        "content_striped",
                        "emoji",
                    ]
                ].dropna(subset="content"),
                "multi",
            )
        if (
            self.data.reactions.count()
            and (SubModules == "All" or "Reacts" in SubModules)
            and not hasattr(self, "Reacts")
        ):
            print("Creating Reacts ...")
            self.Reacts = all_emoji(
                "reactions",
                self.data[["timestamp_ms", "chat", "sender_name", "reactions"]].dropna(),
            )
        if (
            self.data.share.count()
            and (SubModules == "All" or "Links" in SubModules)
            and not hasattr(self, "Links")
        ):
            print("Creating Links ...")
            self.Links = links(
                self.data[["timestamp_ms", "chat", "sender_name", "share"]].dropna(),
                "multi",
            )

        if chat_instances:
            print("Create Chat instances ...")
            subset = False
            if not isinstance(chat_instances, bool):
                if isinstance(chat_instances, str):
                    chat_instances = [chat_instances]
                not_existed_chats = set(chat_instances).difference(
                    set(self.data["chat"].unique())
                )
                if not_existed_chats:
                    raise ValueError(f"{not_existed_chats.pop()} not exist")
                subset = True
            if not len(self.__Chats):
                for name in tqdm(subset and chat_instances or self.data["chat"].unique()):
                    self.__Chats[name] = Chat(
                        name, self.data[self.data["chat"] == name], SubModules
                    )
            else:
                if chat_instances == True:
                    chat_instances = self.data["chat"].unique()
                new_chats = set(chat_instances).difference(set(self.__Chats.keys()))
                if new_chats:
                    print("Adding new chats")
                    for name in tqdm(new_chats):
                        self.__Chats[name] = Chat(
                            name, self.data[self.data["chat"] == name], SubModules
                        )
                    print("Update existing chats")
                for name in tqdm(self.__Chats.keys()):
                    self.__Chats[name]._Chat__reload(SubModules)
                # self.Chats[name]=Chat(name,self.data[self.data["chat"]==name],SubModules)

    def update(self, base, after=True, freeze=True, **kwargs):
        """
        Extend your data.
        Useful for continuous data update, as instead of downloading the entire data each time,
        you can just add the new data to the existing one.

        Parameters
        ----------
        base : str, required
            The directory of the data to be added.
        after : bool, default True
            Whether the new data comes chronologically after the existing data.
            This allows you to add data older than the existing one too .
        freeze : bool, default True
            Whether to freeze the newly created combined data or not.
            Please note that it will override yor existing data.
        **kwargs : parameters to control data generation
            parallel : bool , default False
                When True , data is processed in parallel whenever it possible.
                It helps speed up the process of generating the data, but it consumes your computer resources.
            max_workers : int, default 2
                The number of threads to use when process the data in parallel.
            process_audio : bool, default True
                whether to process the audio files to get their durtion or not.


        Notes
        -----
        For consistency, dur_unit is automatically set to the time unit of the existing data.

        The new data is added after/before the date of the first/last record in the existing data,
        no matter the first/last date of the data to be added.
        `first_msg_date`,`last_msg_date` will help you avoid downloading unnecessary data,
        just download the the data starting from the date of `last_msg_date` or ending at the date of `first_msg_date`.

        if you used `load`, reloading process would be done.
        """

        if not os.path.exists(base):
            raise FileNotFoundError("directory not found")
        if kwargs.get("time_unit"):
            raise ValueError(
                "time_unit will be automatically selected to match the existing data time unit"
            )
        try:
            meta_data = get_meta_data("gen_meta_data.json")
        except:
            raise FileNotFoundError(
                "There is no existing data to be updated, sample data cannot be used, use Stats() to create the data for first time"
            )
        df = df_from_jsons(base, dur_unit=meta_data["time_unit"], **kwargs)
        if after:
            cut_date = self.data.timestamp_ms.max().normalize()
            df1 = self.data[self.data.set_index("timestamp_ms").index < cut_date]
            df2 = df[df.set_index("timestamp_ms").index >= cut_date]
        else:
            cut_date = self.data.timestamp_ms.min().normalize()
            df1 = self.data[self.data.set_index("timestamp_ms").index >= cut_date]
            df2 = df[df.set_index("timestamp_ms").index < cut_date]
        self.data = (
            pd.concat([df1, df2], ignore_index=True)
            .sort_values(["chat", "timestamp_ms"], ascending=[True, False])
            .reset_index(drop=True)
        )
        self.data = self.data.drop(["index"], axis=1, errors="ignore").reset_index()
        if freeze:
            self.freeze()

        if self.__loaded:
            attrs = [i for i in self.__dict__.keys() if i[:1] != "_"]
            for name in attrs:
                delattr(self, name)
            self.__init__(restore=True)
            print("Reloading ...")
            self.load()



import numpy as np
import pandas as pd
import emoji
import matplotlib.pyplot as plt
from imojify import imojify
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from .plot_helper import (
    cat_Vs_counts,
    dynamic_sort,
    make_pretty,
    cat_filter,
    dt_filter,
    freq_mapper,
    offset_image,
    title_freq,
    valid_cum_aggr,
)
from collections import Counter
import arabic_reshaper
from bidi.algorithm import get_display
from pandas.util._decorators import doc


def explode_emoji(txt):
    emojis = emoji.emoji_list(txt)
    return [emojis[i]["emoji"] for i in range(len(emojis))]


def emo_bar(s, n, label=True, **kwargs):
    emjis = s[:n].index
    values = s[:n].values

    fig, ax = plt.subplots(
        figsize=kwargs.get("figsize", (1.25 * n > 15 and 1.25 * n or 15, 10))
    )

    ax.bar(range(len(emjis)), values, width=0.9, align="center")
    ax.set_xticks(range(len(emjis)))
    ax.set_xticklabels([])
    ax.tick_params(axis="x", which="major", pad=26)
    tol = values.max() * 0.02
    for i, e in enumerate(emjis):
        offset_image([i, 0], e, ax, zoom=0.04)
        if label:
            ax.text(
                i,
                values[i] + tol,
                values[i],
                ha="center",
                rotation=kwargs.get("label_rot", 0),
            )
    if label:
        ax.set_ylim(top=tol * 55)
    plt.title(f"Top {n} used emoji combinations", fontsize=12, fontweight="bold")
    return ax


def emoji_filter(df, include=[], exclude=[], keep_var=True, strip_others=False):
    if include and exclude:
        raise Exception(f"include and exclude are mutually exclusive")
    if include:
        if isinstance(include, str):
            include = [include]
        df = df[
            (keep_var and df.emoji.str.contains("|".join(include)))
            | df.emoji.isin(include)
        ]
        s = df.set_index("emoji").squeeze()
        if not strip_others or not keep_var:
            return s

        include_set = set(include)
        s.index = s.index.map(
            lambda txt: [
                emoji["emoji"] if emoji["emoji"] in include_set else ""
                for emoji in emoji.emoji_list(txt)
            ]
        ).map("".join)
        return s.groupby(level=0).sum().sort_values(ascending=False)
    return (
        df.replace(exclude, "", regex=not keep_var)
        .replace("", np.NaN)
        .dropna()
        .set_index("emoji")
        .squeeze()
    )


@doc(
    descr="""The class that enables you to deal with chat Reacts data.
It contains a variety of Methods and Attributes that aim
to help you better understand the data about your chat Reacts.

It handels the Reacts of one single chat.""",
    other_attrs="",
    other_methods="",
    klass="react",
    other="",
)
class chat_emoji:
    """
    {descr}

    {other}Attributes
    ----------
    data : pandas.DataFrame
        The {klass}s row data
    count : int
        The count of all {klass}s
    emoji_list : pandas.Series
        List of all used {klass}s and their counts
    top_emoji : str
        The most used {klass}
    {other_attrs}

    {other}Methods
    -------
    most_used_emojis(...):
        Make bar chart of the most used {klass}s and their counts.
    emoji_timeline(...)
        Make timeline line chart/s for the Top used {klass}s.
    most_used_timeline(...)
        Make line chart showing the most used {klass} for every time period and its count.
    {other_methods}

    """

    def __init__(self, metric, passed_data):
        self._metric = metric
        metric = self._metric == "reactions" and self._metric or "emoji_single"
        self.data = passed_data
        self.count = self.data[metric].count()
        self.emoji_list = self.data[metric].value_counts()
        self.top_emoji = len(self.emoji_list) and self.emoji_list.index[0] or ""

    def most_used_emojis(
        self, n=10, show_sender=True, stacked=False, label=True, legend=True, **kwargs
    ):
        """
        Make bar chart of the most used Rects/Emojis and their counts.

        Parameters
        ----------
        n : int, default 10
            The number of emojis/reacts to show.
        show_sender : bool, default True
            Whether to show emoji/react sender or not.
            When True, 2 bars for each emoji are ploted, one for emojis you sent,
            and one for emojis your friends sent.
        stacked : bool, default False
            If show_sender, whether to stack senders' bars together or not.
        label : bool, default True
            Whether to show the text(annotation) of the bar values (heights).
        legend : bool, default True
            Whether to show the plot legend or not
            Only has effect when `show_sender=True`.
        **kwargs : control the plot appearance
            figsize : tuple (w, h) 
                control the plot width and height
            rot : int , typically an angle between 0 and 360
                change the rotation x axis and y axis tick labels.
            label_rot : int , typically an angle between 0 and 360
                change the rotation of the annotation text.
        Returns
        -------
        matplotlib.axes.Axes
        """

        unexpected_kwargs = set(kwargs.keys()).difference(
            set(["figsize", "rot", "label_rot"])
        )
        if unexpected_kwargs:
            raise TypeError(
                f"got an unexpected keyword argument {unexpected_kwargs.pop()}"
            )

        df = self.data.copy()
        metric = self._metric == "reactions" and self._metric or "emoji_single"
        df, n = cat_filter(df, n, cat=metric, metric="chat")
        return cat_Vs_counts(
            df,
            "chat",
            cat=metric,
            show_sender=show_sender,
            n=n,
            legend=legend,
            label=label,
            stacked=stacked,
            emoji=True,
            **kwargs,
        )

    def emoji_timeline(
        self,
        n=5,
        cumulative=False,
        cum_aggr="sum",
        window="default",
        include=[],
        exclude=[],
        dt_disc=False,
        dt_hrchy="month",
        dt_range="last year",
        subplots=False,
        **kwargs,
    ):
        """
        Make timeline line chart/s for the Top used Rects/Emojis.

        Parameters
        ----------
        n : int, default 5
            The number of emojis/reacts to show.
            Has no effect when `include` parameter is used.
        include : list-like objects, default []
            list of emojis/reacts to include.
        exclude : list-like objects, default []
            list of emojis/reacts to exclude.
        cumulative : bool, default False
            Whether to apply a cumulative aggregation or not.
            Used along with `window` paramter to select a moving or a running aggreagtion.
        cum_aggr : {"count","nunique","sum","mean","max","min","var","std"}, default 'sum'
            The aggregation to be applied for the cumulation.
        window : int or 'default', default 'default'
            Used to define the window within it the aggreagtion is done.
            By default it is the length of the data which calculate running aggregates
            Choose a suitable window length to calculate moving aggregates.
        dt_disc : bool, default False
            when True, datetime is treated as a discrete(categorical) variable
            used along with `dt_hrchy`,`dt_range` to manipulate datetime
        dt_hrchy : {'minute','hour','day','week','month','quarter','year'}, default 'month'
            used to select the hierarchy(frequancy) of the datetime variable
        dt_range : list or a str , default 'last year'
            used to select a range to plot the data within
            pass a list with 2 elements first for the start date, the other for the end
            valid input looks like : ['2020','2022-01-01'], ['2019-01-01','2021-01'],
            ['2022-01-01 01:02:22','2022-02-05 11'], 'last year', 'last 2 weeks', 'first 100 months'
            use None or 'Start'/'End' to select the data from the beginning/ending
            such as ['Start','2022'] ,  ['2010','End'] , [None,'2020-02'], ['Start','End']
            When passing a string, format it like {first/last} then a value and end with a time period
            When ``dt_range=None`` the full time range is selected
        subplots : bool, default False
            used to create subplots, one per emoji/react for every selected emoji/react.
        **kwargs : control the plot appearance
            figsize : tuple (w, h) 
                control the plot width and height
        Returns
        -------
        matplotlib.axes.Axes or np.ndarray of them
        """

        unexpected_kwargs = set(kwargs.keys()).difference(set(["figsize"]))
        if unexpected_kwargs:
            raise TypeError(
                f"got an unexpected keyword argument {unexpected_kwargs.pop()}"
            )

        if dt_hrchy not in freq_mapper.keys():
            raise ValueError(
                f"{dt_hrchy} not a valid datetime frequency; supported frequencies are {list(freq_mapper.keys())}"
            )
        df = self.data.copy()
        metric = self._metric == "reactions" and self._metric or "emoji_single"
        if include or exclude:
            df, n = cat_filter(
                df,
                n=n,
                cat=metric,
                include=include,
                exclude=exclude,
                groups=[],
                others=False,
                asc=False,
                sort="fixed",
                metric="sender_name",
            )
        # if n>df[metric].nunique():
        #   raise
        if dt_range:
            df = dt_filter(df, dt_range)
        groupping_params = {
            True: "timestamp_ms",
            False: pd.Grouper(key="timestamp_ms", freq=freq_mapper[dt_hrchy]),
        }
        if dt_disc:
            df.timestamp_ms = getattr(df.timestamp_ms.dt, dt_hrchy)

        df = df[["timestamp_ms", metric]]
        df = (
            df.groupby(groupping_params[dt_disc])[metric]
            .value_counts()
            .to_frame()
            .rename(columns={metric: "counts"})
            .reset_index()
            .pivot(index="timestamp_ms", columns=metric, values="counts")
            .T
        )
        df = df.sort_index(key=df.sum(1).get, ascending=False)[:n].T.fillna(0)
        if cumulative:
            if cum_aggr not in valid_cum_aggr:
                raise ValueError(
                    f"{cum_aggr} is not a valid aggregation function for cumulation; valid aggr are {valid_cum_aggr}"
                )
            if window == "default":
                window = len(df)
            df = df.rolling(window=window, min_periods=1).agg(cum_aggr)
        if n > len(df.columns):
            n = len(df.columns)
        if len(df.columns) < 2:
            subplots = False
        if subplots:
            fig, axes = plt.subplots(
                nrows=n,
                ncols=1,
                figsize=kwargs.get("figsize", (15, 10)),
                sharex=True,
                sharey=True,
            )
            plt.subplots_adjust(wspace=0.05, hspace=0.05)
        else:
            fig, ax = plt.subplots(
                nrows=1, ncols=1, figsize=kwargs.get("figsize", (15, 10))
            )
        x = df.index
        for i in range(n):
            if subplots:
                ax = axes[i]
            y = df[df.columns[i]]
            image = plt.imread(imojify.get_img_path(y.name))
            image_box = OffsetImage(image, zoom=0.05)
            for x0, y0 in zip(x, y):
                ab = AnnotationBbox(image_box, (x0, y0), frameon=False,)
                ax.add_artist(ab)

            ax.plot(x, y, label="")
        type_title = metric == "reactions" and "reacts" or "emojis"
        cum_subtitle_start = cumulative and "Cumulative" or ""
        cum_subtitle_end = (
            cumulative
            and (
                window == len(df)
                and f"(Running {cum_aggr.upper()})"
                or f"(Moving {cum_aggr.upper()} with window = {window})"
            )
            or ""
        )
        if subplots:
            fig.suptitle(
                f"{cum_subtitle_start} {title_freq[dt_hrchy].title()} Timeline of Top {n} used {type_title} {cum_subtitle_end}",
                y=0.92,
                fontsize=12,
                fontweight="bold",
            )
        else:
            plt.title(
                f"{cum_subtitle_start} {title_freq[dt_hrchy].title()} Timeline of Top {n} used {type_title} {cum_subtitle_end}",
                fontsize=12,
                fontweight="bold",
            )
        return ax

    def most_used_timeline(
        self, dt_disc=False, dt_hrchy="month", dt_range="last year", **kwargs
    ):
        """
        Make a single line chart showing the most used emoji/react
        for every time period and its count.
        
        The data points of the line chart are the count of the most
        used emoji/react in each selected dt_hrchy, and its markers are these emojis.


        Parameters
        ----------
        dt_disc : bool, default False
            when True, datetime is treated as a discrete(categorical) variable
            used along with `dt_hrchy`,`dt_range` to manipulate datetime
        dt_hrchy : {'minute','hour','day','week','month','quarter','year'}, default 'month'
            used to select the hierarchy(frequancy) of the datetime variable
        dt_range : list or a str , default 'last year'
            used to select a range to plot the data within
            pass a list with 2 elements first for the start date, the other for the end
            valid input looks like : ['2020','2022-01-01'], ['2019-01-01','2021-01'],
            ['2022-01-01 01:02:22','2022-02-05 11'], 'last year', 'last 2 weeks', 'first 100 months'
            use None or 'Start'/'End' to select the data from the beginning/ending
            such as ['Start','2022'] ,  ['2010','End'] , [None,'2020-02'], ['Start','End']
            When passing a string, format it like {first/last} then a value and end with a time period
            When ``dt_range=None`` the full time range is selected
        **kwargs : control the plot appearance
            figsize : tuple (w, h) 
                control the plot width and height
        Returns
        -------
        matplotlib.axes.Axes 
        """

        unexpected_kwargs = set(kwargs.keys()).difference(set(["figsize"]))
        if unexpected_kwargs:
            raise TypeError(
                f"got an unexpected keyword argument {unexpected_kwargs.pop()}"
            )

        if dt_hrchy not in freq_mapper.keys():
            raise ValueError(
                f"{dt_hrchy} not a valid datetime frequency; supported frequencies are {list(freq_mapper.keys())}"
            )
        df = self.data.copy()

        metric = self._metric == "reactions" and self._metric or "emoji_single"
        if dt_range:
            df = dt_filter(df, dt_range)
        groupping_params = {
            True: ["timestamp_ms"],
            False: [pd.Grouper(key="timestamp_ms", freq=freq_mapper[dt_hrchy])],
        }
        if dt_disc:
            df.timestamp_ms = getattr(df.timestamp_ms.dt, dt_hrchy)
        df = (
            df.groupby(groupping_params[dt_disc])[metric]
            .agg(lambda x: Counter("".join(x)).most_common(1)[:1])
            .explode()
            .dropna()
            .apply(lambda loc: pd.Series(loc))
            .rename(columns={0: metric, 1: "counts"})
        )
        x = df.index
        y = df["counts"]
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", (15, 10)))

        for x0, y0 in zip(x, y):
            image = plt.imread(imojify.get_img_path((df.loc[x0][metric])))
            image_box = OffsetImage(image, zoom=0.05)
            ab = AnnotationBbox(image_box, (x0, y0), frameon=False,)
            ax.add_artist(ab)
        ax.set_ylim(bottom=0, top=max(y) * 1.1)
        ax.plot(x, y, label="")
        type_title = metric == "reactions" and "react" or "emoji"
        plt.title(
            f"The most used {type_title} over {dt_hrchy}s",
            fontsize=12,
            fontweight="bold",
        )
        return ax


@doc(
    chat_emoji,
    descr="""The class that enables you to deal with all chats Reacts data
It contains a variety of Methods and Attributes that aim
to help you better understand the data about your Reacts across all chats.""",
    other_attrs="",
    other_methods="""emoji_Vs_chat(...)
    Make a bar chart describing your Top reacts across Top chats.
single_emoji_per_chats(emoji,n=10,**kwargs)
    Make a bar chart showing your react across chats.
    """,
    klass="react",
    other="",
)
class all_emoji(chat_emoji):
    def emoji_Vs_chat(
        self,
        n_emoji=5,
        include_emoji=[],
        exclude_emoji=[],
        emoji_groups={},
        n_chats=5,
        include_chat=[],
        exclude_chat=[],
        chat_groups={},
        others=False,
        return_data=False,
        style=False,
        label=True,
        legend=True,
        sort="fixed",
        **kwargs,
    ):
        """
        Make a bar chart describing your Top emojis/reacts across Top chats.

        Parameters
        ----------
        n_emoji : int, default 5
            Max number of emojis/reacts to show.
            Has no effect when include_emoji is used.
        include_emoji : list-like object, default []
            List of the only emojis/reacts to show.
        exclude_emoji : list-like object, default []
            List of emojis to exclude.
        emoji_groups : dict, default {}
            Used to create emoji/reacts groups like: 
            {'😂':'Happy', '😄':'Happy','😓':'Sad', '😢':'Sad' }.
            When used, you must specify the group to which every emoji 
            appears in the plot belong, otherwise it will be labeled as 'Others'.
        n_chats : int, default 5
            Number of chats to show.
            Has no effect when include_chat is used or sort='dynamic'.
        include_chat : list-like object, default []
            List of the only chats to include.
        exclude_chat : list-like object, default []
            List of chats to exclude.
        chat_groups : dict, default {}
            Used to make chat groups the same way as emoji_groups. 
        others : bool, default False
            when True, group any chat/emojis other than the selected n 
            or the included ones and label them as 'Others'
        return_data : bool, default False
            When True, return a CrossTab of the plotting data
        style : bool, default False
            Used with return_data to sytle the returned CrossTab,
            typically, it creates highlighted table.
        label : bool, default True
            whether to show the text(annotation) of the bar values (heights)
        legend : bool, default True
            whether to show the plot legend or not.
        sort : {'fixed', 'dynamic'}, default 'fixed'
            Used to select a 'fixed' `n` chats for all emojis,
            who are The Top people generally,
            or select them 'dynamic'ally , each emoji with its own Top `n` chats.
            Has no effect when include is used.
        **kwargs : control the plot appearance
            figsize : tuple (w, h) 
                control the plot width and height
            rot : int , typically an angle between 0 and 360
                change the rotation x axis and y axis tick labels.
            label_rot : int , typically an angle between 0 and 360
                change the rotation of the annotation text.
        Returns
        -------
        np.ndarray of matplotlib.axes.Axes
        """

        unexpected_kwargs = set(kwargs.keys()).difference(
            set(["figsize", "rot", "label_rot"])
        )
        if unexpected_kwargs:
            raise TypeError(
                f"got an unexpected keyword argument {unexpected_kwargs.pop()}"
            )

        metric = self._metric == "reactions" and self._metric or "emoji_single"

        df = self.data.copy()
        # emoji filter
        df, n_emoji = cat_filter(
            df,
            n=n_emoji,
            cat=metric,
            include=include_emoji,
            exclude=exclude_emoji,
            groups=emoji_groups,
            others=others,
            asc=False,
            sort="fixed",
            metric="sender_name",
        )

        # chat_filter
        df, n = cat_filter(
            df,
            n=n_chats,
            cat="chat",
            include=include_chat,
            exclude=exclude_chat,
            groups=chat_groups,
            others=others,
            sort=sort,
            metric="sender_name",
        )

        df = (
            df.groupby([metric, "chat"])
            .agg({"sender_name": "count"})
            .reset_index()
            .pivot(index=metric, columns="chat", values="sender_name")
        )
        df = df.sort_index(key=df.sum(1).get, ascending=False).fillna(0)
        if return_data:
            if style:
                return df[:n].style.pipe(make_pretty)
            return df
        axes, fig = dynamic_sort(
            df[:n_emoji],
            n=n,
            label=label,
            legend=legend,
            width=3 * n_emoji,
            loc="upper center",
            ncol=len(df.columns),
            bbox_to_anchor=(0.5, 0.95),
            emoji=True,
            **kwargs,
        )
        return axes

    def single_emoji_per_chats(self, emoji, n=10, **kwargs):
        """
        Make a bar chart showing your emoji/react across chats.

        Parameters
        ----------
        emoji : str
            The emoji/react you want to inspect.
        n : int, default 10
            Number of chats to show.
        **kwargs : control the plot appearance
            figsize : tuple (w, h) 
                control the plot width and height
            rot : int , typically an angle between 0 and 360
                change the rotation x axis and y axis tick labels.
            label_rot : int , typically an angle between 0 and 360
                change the rotation of the annotation text.

        Returns
        -------
        matplotlib.axes.Axes
        """
        return self.emoji_Vs_chat(
            include_emoji=[emoji], n_chats=n, others=False, label=True, **kwargs
        )


@doc(
    chat_emoji,
    descr="""The class that enables you to deal with the emojis contained in your chat text messages.
It contains a variety of Methods and Attributes that aim
to help you gain insights about emojis you and your friend used,
whether in the level of a single emoji or a combinations of them.""",
    other_attrs="""emoji_per : float
    The emoji to text ratio of this chat.
emoji_combo : pd.Series
    All of your emoji combos and their count.""",
    other_methods="""most_used_emoji_combo(...)
    Make a bar chart of Top used emoji combinations.
single_emoji_var(emo,return_data=False,label=True,**kwargs)
    Make a bar chart of the counts of each distinct length an emoji.""",
    klass="emoji",
    other="",
)
class emojis(chat_emoji):
    def __init__(self, passed_data):
        self.data = passed_data.join(
            passed_data.emoji.dropna()
            .map(explode_emoji)
            .explode(),  # .replace(["",'️',' '],np.NaN).dropna()
            rsuffix="_single",
            how="inner",
        )[["timestamp_ms", "chat", "sender_name", "emoji", "emoji_single"]].dropna(
            subset="emoji_single"
        )
        super().__init__("emoji", self.data)
        self.emoji_per = (
            len(self.emoji_list)
            and round(
                len("".join(passed_data.emoji.dropna()))
                / len("".join(passed_data.content.dropna()))
                * 100,
                2,
            )
            or 0
        )

        # self.emoji_list=pd.Series([emo['emoji'] for emo in emoji.emoji_list(''.join(self.data.emoji))],dtype='object').value_counts()
        self.emoji_combo = (
            passed_data.groupby("emoji").size().sort_values(ascending=False)
        )

    def most_used_emoji_combo(
        self,
        n=10,
        include=[],
        exclude=[],
        keep_var=True,
        strip_others=False,
        label=True,
        return_data=False,
        **kwargs,
    ):
        """
        Make a bar chart of Top used emoji combinations(emojis used together).

        Parameters
        ----------
        n : int, default 10
            number of emoji combos to show
        include : str or list-like object, default []
            emojis to keep, single emoji & emoji combos are allowed,
            for example: '😂' / '😂❤️' / ['😂','❤️'] / ['😂','😂❤️']
            Note that '😂❤️' not the same as ['😂','❤️'],
            as the '😂❤️' is a pattern that will be mattched,
            while  ['😂','❤️'] isn't
        exclude : str or list-like object, default []
            emojis to exclude
        keep_var : bool, default True
            keep the other variation of the emoji,
            for example, if you included '😂',
            '😂😂' and '😂😂😂'will be keept,
             and for exclude, if you excluded '😂',
            '😂😂' and '😂😂😂' will be kept.
        strip_others : bool, default False
            When include, you can choose between strip any other emoji than
            the included ones from the emoji combo or not.
            for example, if the emoji combo is '😂❤️'
            and you included only ['😂'], '😂❤️' will be keept as it is,
            but if strip_others, '❤️' will be stripped.
        label : bool, default True
            whether to show the text(annotation) of the bar values (heights).
        return data : bool, default False
            return the filtered data with no plots
        **kwargs : control the plot appearance
            figsize : tuple (w, h) 
                control the plot width and height
            label_rot : int , typically an angle between 0 and 360
                change the rotation of the annotation text.

        Returns
        -------
        matplotlib.axes.Axes
        """
        unexpected_kwargs = set(kwargs.keys()).difference(set(["figsize", "label_rot"]))
        if unexpected_kwargs:
            raise TypeError(
                f"got an unexpected keyword argument {unexpected_kwargs.pop()}"
            )
        s = self.emoji_combo.copy()
        if include or exclude:
            s = emoji_filter(
                s.reset_index(),
                include,
                exclude,
                keep_var=keep_var,
                strip_others=strip_others,
            )
        if len(s) < 1:
            raise ValueError("Filters you used resulted in empty data")
        if return_data:
            return s
        ax = emo_bar(s, n=n, label=label, **kwargs)

        return ax

    def single_emoji_var(self, emo, return_data=False, label=True, **kwargs):
        """
        Make a bar chart of the counts of each distinct length an emoji.

        distinct length an emoji means that a single emoji could be sent in different counts,
        for example '😂😂' and '😂😂😂' are varations of '😂' but with different lengths.
        
        So if you called this method on '😂', the number of times a message contained only '😂',
        the number of times a message contained only '😂😂'and so on until the max length of '😂' sent in 
        any message, will be plotted as bars, bar for each different length
        with a bar height represent the count of their occurrence.


        Parameters
        ----------
        emo : str
            The emoji to find its variations.
        return_data : bool, defualt Flase
            Returns the plotting data.
            Could be helpful if the data is too big to be fitted into a single chart.
        label : bool, default True
            Whether to show the text(annotation) of the bar values (heights) 
        **kwargs : control the plot appearance
            figsize : tuple (w, h) 
                control the plot width and height
            label_rot : int , typically an angle between 0 and 360
                change the rotation of the annotation text.

        Returns
        -------
        matplotlib.axes.Axes
        """

        unexpected_kwargs = set(kwargs.keys()).difference(set(["figsize", "label_rot"]))
        if unexpected_kwargs:
            raise TypeError(
                f"got an unexpected keyword argument {unexpected_kwargs.pop()}"
            )

        if not emoji.emoji_list(emo):
            raise ValueError(f"{emo} is not an emoji")
        if not emoji.is_emoji(emo):
            raise ValueError(
                "you may have entered more than one emoji or an emoji combined with text"
            )
        df = (
            self.data[~self.data.index.duplicated(keep="first")]
            .drop(columns=["emoji_single"])
            .copy()
        )
        single_emo = df.emoji.str.extract(f"({emo}+)").dropna()
        if not len(single_emo):
            raise ValueError(f"{emo} not found in your messages")
        single_emo["count"] = single_emo[0].apply(len)
        single_emo = single_emo.sort_values("count").groupby("count").count()
        if return_data:
            return single_emo
        ax = single_emo.plot.bar(
            figsize=kwargs.get(
                "figsize",
                (len(single_emo) * 0.6 > 15 and len(single_emo) * 0.6 or 15, 10),
            ),
            legend=False,
            width=0.9,
        )
        if label:
            tol = single_emo.max().values[0] * 0.02
            for idx, row in single_emo.reset_index().iterrows():
                ax.text(
                    idx,
                    row[0] + tol,
                    row[0],
                    ha="center",
                    rotation=kwargs.get("label_rot", 0),
                )
            ax.set_ylim(top=tol * 55)
        plt.title(
            f"The Count of different lengths of the emoji",
            fontsize=12,
            fontweight="bold",
        )
        return ax


@doc(
    chat_emoji,
    descr="",
    other_attrs="",
    other_methods="",
    klass="emoji",
    other="Other ",
)
class Multi_emojis(emojis, all_emoji):
    """
    The class that enables you to deal with the emojis in all of your messages.
    It contains a variety of Methods and Attributes that aim to help you gain insights
    about all of the emojis used in all chats,
    whether in the level of a single emoji or a combinations of them.

    Attributes
    ----------
    emoji_per : float
        The total emoji to text ratio of all of your chats.
    emoji_combo : pd.Series
        All of your emoji combos and their count.
    most_emoji_ratio : pd.Series
        Emoji to text ratio for each chat, sorted.

    Methods
    -------
    emoji_to_content_ratio(...)
        Make a bar chart of chats with the highest emoji to text ratio.
    most_used_emoji_combo(...)
        Make a bar chart of Top used emoji combinations.
    single_emoji_var(emo,return_data=False,label=True,**kwargs)
        Make a bar chart of the counts of each distinct length an emoji
    emoji_Vs_chat(...)
        Make a bar chart describing your Top emojis across Top chats.
    single_emoji_per_chats(emoji,n=10,**kwargs)
        Make a bar chart showing your emojis across chats.
    """

    def __init__(self, passed_data):
        super().__init__(passed_data)
        self.most_emoji_ratio = (
            passed_data.dropna(subset="content")
            .groupby("chat")[["content", "emoji"]]
            .apply(
                lambda ct: round(
                    len("".join(ct.emoji.dropna()))
                    / len("".join(ct.content.dropna()))
                    * 100,
                    2,
                )
            )
            .sort_values(ascending=False)
        )

    def emoji_to_content_ratio(
        self, n=10, include=[], exclude=[], label=True, **kwargs
    ):
        """
        Make a bar chart of chats with the highest emoji to content(messages text) ratio.

        Parameters
        ----------
        n : int, default 10
            Number of chats to show.
            Has no effect when include is used.
        include : list-like object, default []
            List of the only chats to include.
        exclude : list-like object, default []
            List of chats to exclude
        
        Returns
        -------
        matplotlib.axes.Axes
        """

        unexpected_kwargs = set(kwargs.keys()).difference(
            set(["figsize", "rot", "label_rot"])
        )
        if unexpected_kwargs:
            raise TypeError(
                f"got an unexpected keyword argument {unexpected_kwargs.pop()}"
            )

        s = self.most_emoji_ratio
        if include:
            s = s[s.index.isin(include)]
        if exclude:
            s = s[~s.index.isin[exclude]]
        s = s[:n]
        s.index = s.index.map(lambda name: get_display(arabic_reshaper.reshape(name)))
        ax = s.plot.bar(
            xlabel="",
            figsize=kwargs.get("figsize", (n < 15 and 15 or n, 10)),
            rot=kwargs.get("rot", 45),
            width=0.9,
        )
        if label:
            tol = s.max() * 0.02
            for idx, item in s.reset_index().iterrows():
                ax.text(
                    idx,
                    item.values[1] + tol,
                    round(item.values[1], 2),
                    ha="center",
                    rotation=kwargs.get("label_rot", 0),
                )
            ax.set_ylim(top=tol * 55)
        plt.title(f"emoji-to-word ratio per chat", fontsize=12, fontweight="bold")
        return ax



import numpy as np
import cv2
import os
import json
import pandas as pd
import librosa
from multiprocessing.pool import ThreadPool as Pool
import emoji
from tqdm import tqdm
import math
import regex

RE_BAD_CHARS = regex.compile(r"[\p{Cc}\p{Cs}]+")


def remove_bad_chars(text):
    return RE_BAD_CHARS.sub("", text)


time_units = ["sec", "minute", "hour", "day"]
unit_factors = [60, 60, 24]


def get_unit_factor(cur_unit, unit):
    if unit not in time_units:
        raise ValueError(
            f"{unit} is not a valid time unit; valid units are {time_units}"
        )
    indces = [time_units.index(cur_unit), time_units.index(unit)]
    factor = math.prod(unit_factors[min(indces) : max(indces)])
    return factor, indces


def set_dur_unit(data, cur_unit, unit):

    factor, indces = get_unit_factor(cur_unit, unit)
    if indces[1] > indces[0]:
        data = data.map(lambda dur: round(dur / factor, 2))
    else:
        data = data.map(lambda dur: round(dur * factor, 2))
    return data


def get_video_dur(filename):
    try:
        cap = cv2.VideoCapture(filename)
        fps = cap.get(cv2.CAP_PROP_FPS)
        frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        duration = fps > 0 and round(frame_count / fps, 2) or 0
        cap.release()
        return duration
    except:
        return 0


def get_audio_length(uri):
    try:
        return librosa.get_duration(filename=uri)
    except:
        return 0


def parallel_audio(base, folders, max_workers=2):

    directory = base + "messages/inbox"
    s = (
        pd.Series(
            list(
                map(
                    lambda direc: os.path.isdir(direc)
                    and list(map(lambda txt: direc + "/" + txt, os.listdir(direc)))
                    or np.NaN,
                    list(map(lambda txt: directory + "/" + txt + "/audio", folders)),
                )
            )
        )
        .dropna()
        .explode()
    )

    # folders = os.listdir(directory)
    results = []

    with Pool(processes=max_workers) as p:
        max_ = len(s)
        with tqdm(total=max_) as pbar:
            for _ in p.imap(get_audio_length, s):
                results.append(_)
                pbar.update()

    s.index = results
    s = s.str.replace(base, "", regex=False)
    s = s = pd.Series(s.index, index=s.values)
    return s


def parse_json(base, parallel=False, process_audio=True, dur_unit="sec", **kwargs):

    jsons = {}
    base = base.replace("\\", "/")
    if base[-1] != "/":
        base += "/"
    directory = base + "messages/inbox"

    folders = os.listdir(directory)
    info = json.load(
        open(os.path.join(base, "messages\\autofill_information.json"), "r")
    )
    user = info[list(info)[-1]]["FULL_NAME"][0]
    if ".DS_Store" in folders:
        folders.remove(".DS_Store")
    unit_factor = (
        (dur_unit == "sec" or parallel) and 1 or get_unit_factor("sec", dur_unit)[0]
    )
    if parallel and process_audio:
        print("Getting Audio Files Duration In Parallel ...")
        max_workers = kwargs.get("max_workers", 2)
        s = parallel_audio(base, folders, max_workers)
        if dur_unit != "sec":
            s = set_dur_unit(s, "sec", dur_unit)

    print("Parsing Data ...")
    for folder in tqdm(folders):
        for filename in os.listdir(os.path.join(directory, folder)):
            if filename.startswith("message"):
                data = json.load(open(os.path.join(directory, folder, filename), "r"))

                if len(data["participants"]) != 2:
                    continue
                if data["participants"][0]["name"] != user:
                    data["participants"] = {"name": data["participants"][0]["name"]}
                else:
                    data["participants"] = {"name": data["participants"][1]["name"]}
                if '"' in data["participants"]["name"]:
                    data["participants"]["name"] = data["participants"]["name"].replace(
                        '"', ""
                    )
                data["participants"]["name"] = r'"{}"'.format(
                    data["participants"]["name"]
                )
                data["participants"]["name"] = (
                    json.loads(data["participants"]["name"])
                    .encode("latin1")
                    .decode("utf8")
                )

                for indx, message in enumerate(data["messages"]):

                    if "audio_files" in data["messages"][indx].keys():
                        for aud_idx, audio in enumerate(
                            data["messages"][indx]["audio_files"]
                        ):
                            dirc = (
                                base
                                + data["messages"][indx]["audio_files"][aud_idx]["uri"]
                            )
                            try:
                                data["messages"][indx]["audio_files"][aud_idx][
                                    "length"
                                ] = (
                                    process_audio
                                    and (
                                        parallel
                                        and s[
                                            data["messages"][indx]["audio_files"][
                                                aud_idx
                                            ]["uri"]
                                        ]
                                        or librosa.get_duration(filename=dirc)
                                    )
                                    or 0
                                )
                            except:
                                data["messages"][indx]["audio_files"][aud_idx][
                                    "length"
                                ] = 0
                            if unit_factor != 1:
                                data["messages"][indx]["audio_files"][aud_idx][
                                    "length"
                                ] = round(
                                    data["messages"][indx]["audio_files"][aud_idx][
                                        "length"
                                    ]
                                    / unit_factor,
                                    2,
                                )
                    if "videos" in data["messages"][indx].keys():
                        for vid_idx, vid in enumerate(data["messages"][indx]["videos"]):
                            dirc = (
                                base + data["messages"][indx]["videos"][vid_idx]["uri"]
                            )
                            data["messages"][indx]["videos"][vid_idx][
                                "duration"
                            ] = get_video_dur(dirc)
                            if unit_factor != 1:
                                data["messages"][indx]["videos"][vid_idx][
                                    "duration"
                                ] = round(
                                    data["messages"][indx]["videos"][vid_idx][
                                        "duration"
                                    ]
                                    / unit_factor,
                                    2,
                                )

                    if "reactions" in data["messages"][indx].keys():
                        data["messages"][indx]["reactions"] = (
                            message["reactions"][0]["reaction"]
                            .encode("latin1")
                            .decode("utf8")
                        )

                    if "content" in data["messages"][indx].keys():
                        data["messages"][indx]["content"] = (
                            message["content"].encode("latin1").decode("utf8")
                        )

                    if "sender_name" in data["messages"][indx].keys():
                        data["messages"][indx]["sender_name"] = (
                            message["sender_name"].encode("latin1").decode("utf8")
                        )

                if data["participants"]["name"] in jsons.keys():

                    jsons[data["participants"]["name"]]["messages"] += data["messages"]
                else:

                    jsons[data["participants"]["name"]] = data
    return jsons


def df_from_jsons(base, parallel=False, process_audio=True, dur_unit="sec", **kwargs):

    tqdm.pandas()

    jsons = parse_json(base, parallel, process_audio, dur_unit, **kwargs)
    print("Building df ...")
    dflist = []
    for chat, json in jsons.items():
        temp = pd.DataFrame(json["messages"])
        temp["chat"] = chat
        dflist.append(temp)
    df = pd.concat(dflist).reset_index(drop=True)
    # df.timestamp_ms=df.timestamp_ms.apply(pd.to_datetime,unit='ms')
    df["call_status"] = df["call_duration"].map(
        lambda val: val > 0 and "replied" or "missed", na_action='ignore'
    )
    xemojis = (
        df.content.dropna()
        .apply(lambda txt: "".join([emo["emoji"] for emo in emoji.emoji_list(txt)]))
        .reset_index()
    )
    xemojis.columns = ["index", "emoji"]
    df = df.reset_index().merge(xemojis, how="left", on="index").replace("", np.NaN)
    df = df.join(
        df.dropna(subset="call_duration")["content"]
        .map(lambda txt: any(video_kw in txt for video_kw in ['فيديو','video']) and "video" or "voice")
        .rename("call_type"),
        how="left",
    )
    df.content = df.content.map(
        lambda txt: remove_bad_chars(txt) ,na_action='ignore'
    )
    if dur_unit != "sec":
        df.call_duration = set_dur_unit(df.call_duration, "sec", dur_unit)
    print("Convert Timestamps to DateTime format ...")
    df.timestamp_ms = df.timestamp_ms.progress_apply(pd.to_datetime, unit="ms")

    return df[
        [
            "index",
            "chat",
            "sender_name",
            "timestamp_ms",
            "content",
            "is_unsent",
            "share",
            "photos",
            "reactions",
            "call_duration",
            "audio_files",
            "gifs",
            "videos",
            "sticker",
            "files",
            "call_status",
            "emoji",
            "call_type",
        ]
    ]



import pandas as pd
import numpy as np
import os
import emoji
import matplotlib.pyplot as plt
import arabic_reshaper
from bidi.algorithm import get_display
import matplotlib.patches as mpatches
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from imojify import imojify

cwd = os.path.dirname(__file__)
import json


def get_meta_data(filename="meta_data.json"):
    try:
        with open(os.path.join(cwd, filename), "r") as openfile:

            # Reading from json file
            meta_data = json.load(openfile)
    except:
        meta_data = {}
    return meta_data


def offset_image(coords, name, ax, zoom=0.04):
    img = plt.imread(imojify.get_img_path(name))
    im = OffsetImage(img, zoom=zoom)
    im.image.axes = ax
    ab = AnnotationBbox(
        im,
        (coords[0], coords[1]),
        xybox=(0.0, -16.0),
        frameon=False,
        xycoords="data",
        boxcoords="offset points",
        pad=0,
    )

    ax.add_artist(ab)


stck_dflts = {
    "count": True,
    "nunique": True,
    "sum": True,
    "mean": False,
    "max": False,
    "min": False,
    "first": False,
    "last": False,
    "var": False,
    "std": False,
}
sort_ops = {True: "sum", False: "mean"}
freq_mapper = {
    "day": "D",
    "week": "W",
    "month": "MS",
    "quarter": "Q",
    "year": "YS",
    "hour": "H",
    "minute": "T",
}
colors = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]
nxtlvlofdtils = {"year": "month", "quarter": "month", "month": "day", "week": "day"}
valid_dt = [
    "years",
    "months",
    "weeks",
    "days",
    "hours",
    "minutes",
    "seconds",
    "microseconds",
    "nanoseconds",
]
metric_categories = {
    "photos": ["chat"],
    "gifs": ["chat"],
    "files": ["chat", "ext"],
    "sticker": ["chat"],
    "audio_files": ["chat"],
    "videos": ["chat"],
    "media_type": ["chat", "media"],
    "content": ["chat"],
    "call_duration": ["chat"],
    "call_type": ["chat"],
    "share": ["chat", "domain"],
    "chat": ["sender_name"],
    "index": ["chat"],
}
title_freq = {
    "year": "Yearly",
    "quarter": "Quarterly",
    "month": "Monthly",
    "week": "Weekly",
    "day": "Daily",
    "hour": "Hourly",
    "minute": "Minutely",
}
metric_display = {
    "media_type": "All Media",
    "index": "Messages",
    "audio_files_uri": "Audio Files",
    "audio_files_length": "Audio duration",
    "call_type": "Calls",
    "call_duration": "Call duration",
    "share": "Shared links",
    "content": "Text messages",
    "chat": "Chat messages",
    "video_uri": "Videos",
    "video_length": "Video duration",
    "files": "Files",
    "sticker": "Stickers",
    "gifs": "Gifs",
    "photos": "Photos",
}
cat_display = {
    "chat": "Chats",
    "media": "Meida types",
    "ext": "File extensions",
    "domain": "Domains",
    "sender_name": "Senders",
    "call_type": "Call type ",
    "call_status": "Call status ",
}
valid_cum_aggr = ["count", "nunique", "sum", "mean", "max", "min", "var", "std"]


def make_pretty(styler):
    styler.background_gradient(axis=None, vmin=1, cmap="Blues")
    styler.set_properties(**{"width": "10em", "text-align": "center"}).set_table_styles(
        [{"selector": "th", "props": [("text-align", "center")]}]
    )
    return styler


def metric_correct(metric, aggr):
    if callable(aggr):
        if metric in ["audio_files", "video"]:
            return metric + "_length"
        return metric
    if aggr not in stck_dflts.keys():
        raise ValueError(
            f"{aggr} is not a valid aggregation function; supported aggr are {list(stck_dflts.keys())}"
        )
    if metric in ["audio_files", "video"]:
        if aggr in ["count", "nunique"]:
            metric = metric + "_uri"
        else:
            metric = metric + "_length"
    elif metric == "call_duration":
        if aggr in ["count", "nunique"]:
            metric = "call_type"
    if aggr not in ["count", "nunique"] and metric not in [
        "call_duration",
        "audio_files_length",
        "video_length",
    ]:
        raise ValueError(
            f"{aggr} is not a supported value for {metric}; supported aggregation are 'count' and 'nunique' "
        )
    return metric


# edit for diffrent agg than count
def cat_filter(
    df,
    n=5,
    cat="chat",
    include=[],
    exclude=[],
    others=False,
    groups={},
    asc=False,
    sort="fixed",
    metric="",
    aggr="count",
):
    df = df.copy()
    # print(f'include :{include},exclude : {exclude},others : {others},cat :{cat}')
    if include:

        if include == "Top" and cat == "chat":
            meta_data = get_meta_data()
            include = meta_data["Top"]
        elif np.ndim(include) == 0:
            raise TypeError(f"{include} is not list-like object")
        if others:
            df[cat] = df[cat].apply(lambda name: name if name in include else "Others")
            n = df[cat].nunique() + 1

        else:
            df = df[df[cat].isin(include)]
            n = df[cat].nunique()
        if n == 0:
            raise ValueError(f"{cat} do not contain any of {include} ")

    else:

        if n in ["All", 0, None]:
            n = df[cat].nunique()
        elif not isinstance(n, int):
            raise TypeError(f"expected an integer, got{type(n)}")
        elif n and sort == "fixed":
            if n < 0:
                n = df[cat].nunique() + n
            top_n = (
                df.groupby(cat)[metric].agg(aggr).sort_values(ascending=asc).index[:n]
            )
            if n > len(top_n):
                n = len(top_n)
            if others:
                df[cat] = df[cat].apply(
                    lambda name: name if name in top_n else "Others"
                )
                n = n + 1
            else:
                df = df[df[cat].isin(top_n)]

    if exclude:
        if np.ndim(exclude) == 0:
            raise TypeError(f"{exclude} is not list-like object")

        df = df[~df[cat].isin(exclude)]
        n = df[cat].nunique()
        if n == 0:
            raise ValueError(f"{cat} do not contain any of {include} ")

    if groups:
        if not isinstance(groups, dict):
            raise TypeError(f"{groups} is not a dict ")
        df[cat] = df[cat].map(lambda name: groups.get(name, "Others"))
        n = df[cat].nunique()
    return df, n


def dt_filter(df, dt_range):

    before = df.timestamp_ms.min()
    after = df.timestamp_ms.max()
    if isinstance(dt_range, str):
        tokens = dt_range.split()
        if len(tokens) < 2:
            raise Exception(f"{dt_range} is not a valid dt_range")
        if tokens[0].lower() in ["first", "last"]:
            if len(tokens) == 2:
                tokens[1] = tokens[1].endswith("s") and tokens[1] or tokens[1] + "s"
                if tokens[1] not in valid_dt:
                    raise Exception(
                        f"{tokens[1]} is not a valid freq, valid freqs : {valid_dt}"
                    )
                DateOffset = pd.DateOffset(**{tokens[1]: 1})
            elif len(tokens) == 3:
                if not tokens[1].isnumeric():
                    raise ValueError(f"{tokens[1]} is not a number")
                d_val = tokens[1]
                tokens[2] = tokens[2].endswith("s") and tokens[2] or tokens[2] + "s"
                if tokens[2] not in valid_dt:
                    raise Exception(
                        f"{tokens[2]} is not a valid freq, valid freqs : {valid_dt}"
                    )
                DateOffset = pd.DateOffset(**{tokens[2]: int(d_val)})
            else:
                raise Exception(f"{dt_range} is not a valid dt_range")
            if tokens[0].lower() == "first":
                first_d = before
                dt_range = [first_d, first_d + DateOffset]
            else:
                last_d = after
                dt_range = [last_d - DateOffset, last_d]
        else:
            raise Exception(
                "A valid string should starts with either 'first' or 'last' "
            )

    if dt_range[0] and dt_range[0] != "Start":
        try:
            before = pd.to_datetime(dt_range[0])
        except:
            raise ValueError(f"{dt_range[0]} is not a valid datetime format")
    if dt_range[1] and dt_range[1] != "End":
        try:
            after = pd.to_datetime(dt_range[1])
        except:
            raise ValueError(f"{dt_range[1]} is not a valid datetime format")
    df = df[(df.timestamp_ms >= before) & (df.timestamp_ms <= after)]
    if not len(df):
        raise ValueError("not data to plot for the date range you specified")
    return df


def label_counts(ax, df, stacked, show_sender, **kwargs):

    if stacked:
        for idx, row in df.reset_index().iterrows():
            if row[df.columns[0]]:
                ax.text(
                    idx,
                    row[df.columns[0]] / 2,
                    row[df.columns[0]],
                    rotation=kwargs.get("label_rot", 90),
                    fontsize=10,
                    ha="center",
                )
            try:
                if row[df.columns[1]]:
                    ax.text(
                        idx,
                        row[df.columns[0]] + (row[df.columns[1]]) / 2,
                        row[df.columns[1]],
                        rotation=kwargs.get("label_rot", 90),
                        fontsize=10,
                        ha="center",
                    )
            except:
                pass
        ax.set_ylim(top=df.sum(axis=1).max() * 1.1)
    else:
        tol = df.max().max() * 0.02
        if not show_sender:

            for idx, row in df.reset_index().iterrows():
                if row[1]:
                    ax.text(
                        idx,
                        row[1] + tol,
                        row[1],
                        rotation=kwargs.get("label_rot", 0),
                        fontsize=10,
                        ha="center",
                    )
        else:

            for idx, row in df.reset_index().iterrows():
                if row[df.columns[0]]:
                    ax.text(
                        "emoji" in kwargs.keys() and idx - 0.2125 or idx - 0.125,
                        row[df.columns[0]] + tol,
                        row[df.columns[0]],
                        rotation=kwargs.get("label_rot", 90),
                        fontsize=10,
                        ha="center",
                    )
                try:
                    if row[df.columns[1]]:
                        ax.text(
                            "emoji" in kwargs.keys() and idx + 0.2125 or idx + 0.125,
                            row[df.columns[1]] + tol,
                            row[df.columns[1]],
                            rotation=kwargs.get("label_rot", 90),
                            fontsize=10,
                            ha="center",
                        )
                except:
                    pass
        ax.set_ylim(top=tol * 55)
    # plt.xticks(rotation=45);
    return ax


def top_timely(
    data,
    metric,
    aggr,
    cat,
    n=20,
    dt_hrchy="month",
    dt_range="last year",
    show_name=False,
    legend=True,
    rank=False,
    **kwargs,
):

    df = data.copy()
    if dt_range:
        df = dt_filter(data, dt_range)
    df = (
        df.groupby([pd.Grouper(key="timestamp_ms", freq=freq_mapper[dt_hrchy]), cat])
        .agg({metric: aggr})
        .reset_index()
    )
    df = (
        df.groupby("timestamp_ms")
        .agg({metric: "max"})
        .reset_index()
        .merge(on=["timestamp_ms", metric], right=df)
    )
    if dt_hrchy in ["hour", "minute"]:
        df.timestamp_ms = df.timestamp_ms.dt.floor(freq_mapper[dt_hrchy])
    else:
        df.timestamp_ms = df.timestamp_ms.dt.date
    if rank:
        df.sort_values(metric, inplace=True, ascending=True)
    if n == None or n == "All":
        n = len(df)
    df = df[-n:].pivot(index="timestamp_ms", columns=cat, values=metric)
    df.columns = df.columns.map(lambda name: get_display(arabic_reshaper.reshape(name)))
    if not rank:
        df = df[-n:].dropna(thresh=1, axis=1)
        ax = df.plot.bar(
            figsize=kwargs.get("figsize", (0.6 * n >= 15 and 0.6 * n or 15, 10)),
            width=0.95,
            stacked=True,
            xlabel="",
            legend=legend,
        )
        plt.xticks(rotation=kwargs.get("rot", 45))
        tol = df.sum(1).max() * 0.01

        for row in df.reset_index().iterrows():
            name = row[1].dropna().keys().values[1:]
            val = row[1].dropna()[name[:1]]
            if show_name:
                ax.text(
                    row[0] + 0.2,
                    tol,
                    ", ".join(name),
                    color="black",
                    fontsize=10,
                    rotation=kwargs.get("label_rot", 90),
                    ha="center",
                    va="bottom",
                )
                ax.text(
                    row[0] - 0.2,
                    tol,
                    int(val),
                    color="black",
                    fontsize=10,
                    rotation=kwargs.get("label_rot", 90),
                    ha="center",
                    va="bottom",
                )
            else:
                ax.text(
                    row[0],
                    val + tol,
                    int(val),
                    color="blue",
                    fontsize=10,
                    ha="center",
                    rotation=kwargs.get("label_rot", 0),
                )
        ax.set_ylim(top=df.sum(1).max() * 1.1)
        plt.ylabel(f"COUNT OF {metric_display[metric].upper()}", fontdict={"size": 12})

        plt.title(
            f"Highest {title_freq[dt_hrchy]} {aggr.title()} of {metric_display[metric]} By {cat_display[cat][:-1]} -Timeline",
            fontdict={"weight": "bold", "size": 12},
        )
        if legend:
            plt.legend(ncol=len(df.columns), loc="upper center")
    else:
        df = df.sort_index(key=df.sum(1).get)
        ax = df.plot.barh(
            stacked=True,
            legend=legend,
            figsize=kwargs.get("figsize", (15, 0.6 * n >= 10 and 0.6 * n or 10)),
            ylabel="",
            xlabel="",
            width=0.9,
            rot=kwargs.get("rot"),
        )
        tol = df.sum(1).max() * 0.005
        for row in df.reset_index().iterrows():
            name = row[1].dropna().keys().values[1:]
            val = row[1].dropna()[name[:1]]
            ax.text(
                val + tol,
                row[0] + 0.25,
                ", ".join(name),
                color="blue",
                fontsize=10,
                ha="left",
                va="center",
                rotation=kwargs.get("label_rot", 0),
            )
            ax.text(
                val + tol,
                row[0] - 0.25,
                int(val),
                color="blue",
                fontsize=10,
                ha="left",
                va="center",
                rotation=kwargs.get("label_rot", 0),
            )
        ax.set_xlim(right=df.sum(1).max() * 1.15)
        # plt.legend(prop={'weight':'bold'})
        plt.xlabel(f"COUNT OF {metric_display[metric].upper()}", fontdict={"size": 12})
        plt.title(
            f"Highest {title_freq[dt_hrchy]} {aggr.title()} of {metric_display[metric]} By {cat_display[cat]}(Ranked)",
            fontdict={"weight": "bold", "size": 12},
        )

        # plt.title(f'Ranked Most {title_freq[dt_hrchy] } {metric_display[metric]} By {cat.title()}',fontdict={'weight':'bold','size':12})
        ax.yaxis.set_ticks_position("none")
    # plt.yticks(weight = 'bold')
    # plt.xticks(weight = 'bold')

    return ax


def sum_of_most(data, metric, aggr, cat, n=10, dt_hrchy="day", **kwargs):
    df = data.copy()
    df = (
        df.groupby([pd.Grouper(key="timestamp_ms", freq=freq_mapper[dt_hrchy]), cat])
        .agg({metric: aggr})
        .reset_index()
    )
    df = (
        df.groupby("timestamp_ms")
        .agg({metric: "max"})
        .reset_index()
        .merge(on=["timestamp_ms", metric], right=df)
    )
    df.sort_values(metric, inplace=True, ascending=True)
    df = pd.DataFrame(df[[cat]].value_counts()[:n]).reset_index()
    df[cat] = df[cat].map(lambda name: get_display(arabic_reshaper.reshape(name)))
    ax = df.plot.bar(
        x=cat,
        y=0,
        color=plt.cm.tab20(range(n)),
        figsize=kwargs.get("figsize", (n > 15 and n or 15, 10)),
        legend=False,
        xlabel="",
        rot=kwargs.get("rot", 45),
        width=0.9,
    )
    tol = df[0].max() * 0.02
    for row in df.iterrows():

        val = row[1][0]
        if val:
            ax.text(
                row[0],
                val + tol,
                int(val),
                color="blue",
                fontsize=10,
                ha="center",
                va="center",
                rotation=kwargs.get("label_rot", 0),
            )

    plt.title(
        f"The Total Number of Times a {cat_display[cat][:-1]} Had The Highest {title_freq[dt_hrchy]} {aggr.title()} of {metric_display[metric]}",
        fontdict={"weight": "bold", "size": 12},
    )
    # The total number of times a chat had the highest daily count of messages
    return ax


def most_per_chat(data, metric, aggr, n=10, dt_hrchy="day", **kwargs):
    df = data.copy()
    df = (
        df.groupby([pd.Grouper(key="timestamp_ms", freq=freq_mapper[dt_hrchy]), "chat"])
        .agg({metric: aggr})
        .reset_index()
    )
    df = (
        df.groupby("chat")[[metric]]
        .max()
        .sort_values(metric, ascending=False)[:n]
        .reset_index()
    )
    df.chat = df.chat.map(lambda name: get_display(arabic_reshaper.reshape(name)))
    # return df
    ax = df.plot.bar(
        x="chat",
        y=metric,
        color=plt.cm.tab20(range(n)),
        figsize=kwargs.get("figsize", (n > 15 and n or 15, 10)),
        legend=False,
        xlabel="",
        rot=kwargs.get("rot", 45),
        width=0.9,
    )
    tol = df[metric].max() * 0.02
    for row in df.iterrows():
        val = row[1][metric]
        if val:
            ax.text(
                row[0],
                val + tol,
                int(val),
                color="blue",
                fontsize=10,
                ha="center",
                va="center",
                rotation=kwargs.get("label_rot", 0),
            )
    plt.title(
        f"Most #Of {metric_display[metric]} Sent In One {dt_hrchy.title()} Per Chat",
        fontdict={"weight": "bold", "size": 12},
    )
    plt.ylabel(f"COUNT OF {metric_display[metric].upper()}", fontdict={"size": 12})
    return ax


def dt_Vs_counts(
    df,
    metric,
    aggr="count",
    cumulative=False,
    cum_aggr="sum",
    window="default",
    dt_disc=False,
    dt_hrchy="month",
    dt_groups={},
    show_sender=True,
):
    if dt_hrchy not in freq_mapper.keys():
        raise ValueError(
            f"{dt_hrchy} not a valid datetime frequency; supported frequencies are {list(freq_mapper.keys())}"
        )

    meta_data = get_meta_data()

    groupping_params = {
        True: ["timestamp_ms", "sender_name"],
        False: [
            pd.Grouper(key="timestamp_ms", freq=freq_mapper[dt_hrchy]),
            "sender_name",
        ],
    }
    if df.sender_name.nunique() > 2:
        df["sender_name"] = df["sender_name"].apply(
            lambda name: name if name in [meta_data["user"]] else "Others"
        )
    if dt_disc:

        df.timestamp_ms = getattr(df.timestamp_ms.dt, dt_hrchy)
        if dt_groups:
            df.timestamp_ms = df.timestamp_ms.map(
                lambda dt: dt_groups.get(dt, "Others")
            )

    if show_sender:
        df = (
            df.groupby(groupping_params[dt_disc])
            .agg({metric: aggr})
            .reset_index()
            .pivot(index="timestamp_ms", columns="sender_name", values=metric)
        )
    else:
        df = df.groupby(groupping_params[dt_disc][0]).agg({metric: aggr})

    if not dt_disc:
        df = df.reindex(
            pd.date_range(df.index[0], df.index[-1], freq=freq_mapper[dt_hrchy])
        ).fillna(0)
        df.index = df.index.date
    df = df.astype("int")
    if cumulative:
        if cum_aggr not in valid_cum_aggr:
            raise ValueError(
                f"{cum_aggr} is not a valid aggregation function for cumulation; valid aggr are {valid_cum_aggr}"
            )
        if window == "default":
            window = len(df)

        df = df.rolling(window=window, min_periods=1).agg(cum_aggr)
        df = df.apply(round, args=[2])
    return df


def dt_Vs_counts_bar(
    df,
    metric,
    aggr="count",
    cumulative=False,
    cum_aggr="sum",
    window="default",
    dt_disc=False,
    dt_hrchy="month",
    dt_groups={},
    show_sender=True,
    label=True,
    stacked=False,
    return_data=False,
    style=False,
    legend=True,
    **kwargs,
):
    df = dt_Vs_counts(
        df,
        metric,
        aggr,
        cumulative,
        cum_aggr,
        window,
        dt_disc,
        dt_hrchy,
        dt_groups,
        show_sender,
    )

    if not show_sender:
        stacked = False
        legend = False
    if return_data:
        if style:
            return df.style.pipe(make_pretty)
        return df

    df.columns = df.columns.map(lambda name: get_display(arabic_reshaper.reshape(name)))
    # df.index=df.index.map(lambda name:get_display(arabic_reshaper.reshape(name)))
    ax = df.plot.bar(
        figsize=kwargs.get("figsize", (len(df) // 10 and len(df) or 10, 10)),
        stacked=stacked,
        legend=legend,
        xlabel="",
    )
    if label:
        ax = label_counts(ax, df, stacked, show_sender, **kwargs)
    plt.xticks(rotation=kwargs.get("rot", 45))
    if window == "default":
        window = len(df)
    cum_subtitle_start = cumulative and "Cumulative" or ""
    cum_subtitle_end = (
        cumulative
        and (
            window == len(df)
            and f"(Running {cum_aggr.upper()})"
            or f"(Moving {cum_aggr.upper()} with window = {window})"
        )
        or ""
    )
    plt.title(
        f"{cum_subtitle_start} {title_freq[dt_hrchy]} Timeline of the {aggr} of {metric_display[metric]} {cum_subtitle_end}",
        fontsize=12,
        fontweight="bold",
    )
    plt.ylabel(f"{aggr} of {metric_display[metric]}".upper(), fontdict={"size": 12})
    return ax


def cat_Vs_counts(
    df,
    metric,
    show_sender=False,
    cat="chat",
    n=5,
    stacked=False,
    label=True,
    legend=True,
    asc=False,
    aggr="count",
    return_data=False,
    style=True,
    *args,
    **kwargs,
):

    if df.sender_name.nunique() > 2:
        meta_data = get_meta_data()
        df["sender_name"] = df["sender_name"].apply(
            lambda name: name if name in [meta_data["user"]] else "Others"
        )

    if show_sender:
        df = (
            df.groupby([cat, "sender_name"])
            .agg({metric: aggr})
            .reset_index()
            .pivot(index=cat, columns="sender_name", values=metric)
        )
        df = df.fillna(0).sort_index(key=df.agg("sum", axis=1).get, ascending=asc)
        df = df.astype("int")
    else:

        df = df[cat].value_counts(ascending=asc).to_frame()
        stacked = False
        legend = False
    if "emoji" in kwargs.keys() and cat == "reactions" and show_sender:
        try:
            df.columns = [df.columns[1], df.columns[0]]
            df = df[[df.columns[1], df.columns[0]]]
        except:
            stacked = True
    if return_data:
        if style:
            return df.style.pipe(make_pretty)
        return df
    if "emoji" not in kwargs.keys():
        df.index = df.index.map(lambda name: get_display(arabic_reshaper.reshape(name)))
        df.columns = df.columns.map(
            lambda name: get_display(arabic_reshaper.reshape(name))
        )
    # ax=df.plot.bar(figsize=(len(df.columns)>15 and len(df) or 15,10),stacked=stacked,xlabel="",ylabel=f"{aggr} of {metric}".upper(),legend=legend)
    ax = df.plot.bar(
        figsize=kwargs.get(
            "figsize", "emoji" in kwargs.keys() and (len(df) > 15 and len(df) or 15, 10)
        )
        or (len(df.columns) // 10 and len(df) or 10, 10),
        stacked=stacked,
        xlabel="",
        legend=legend,
        width="emoji" in kwargs.keys() and 0.85 or 0.5,
    )
    if "emoji" in kwargs.keys():
        for idx, emo in enumerate(df.index):
            offset_image([idx, 0], emo, ax)
        ax.xaxis.set_ticklabels([])
        sub_title = cat == "reactions" and "react" or "absolute emoji"
        plt.title(f"Top {n} used {sub_title}", fontsize=12, fontweight="bold")
    else:
        plt.title(
            f"The {n} {cat_display[cat]} with the highest {aggr} of {metric_display[metric]}",
            fontsize=12,
            fontweight="bold",
        )
    if label:
        ax = label_counts(ax, df, stacked, show_sender, **kwargs)
    plt.xticks(rotation=kwargs.get("rot", 45))
    plt.ylabel(f"{aggr} of {metric_display[metric]}".upper(), fontdict={"size": 12})
    return ax


def dynamic_sort(df, n=5, label=True, legend=True, width=15, *args, **kwargs):
    df.columns = df.columns.map(lambda name: get_display(arabic_reshaper.reshape(name)))
    if len(df) < 2:
        df = df.T.sort_index(key=df.T.sum(1).get, ascending=False)[:n]
        ax = df.plot.bar(
            figsize=kwargs.get("figsize", (n < 15 and 15 or n, 10)),
            legend=False,
            xlabel="",
            rot=kwargs.get("rot", 45),
            width=0.9,
        )
        if "emoji" in kwargs.keys():
            plt.title(
                f"The use of Emoji across chats", y=1, fontsize=12, fontweight="bold"
            )
        if label:
            tol = df.max().values[0] * 0.02
            for idx, row in df.reset_index().iterrows():
                if row[1]:
                    ax.text(
                        idx,
                        row[1] + tol,
                        row[1],
                        ha="center",
                        rotation=kwargs.get("label_rot", 0),
                    )
            ax.set_ylim(top=tol * 55)
        return ax, ax.get_figure()
    chat_colors = {}
    handles = []
    fig, axes = plt.subplots(
        nrows=1, ncols=len(df), sharey=True, figsize=kwargs.get("figsize", (width, 10))
    )
    plt.subplots_adjust(wspace=0, hspace=0)
    tol = df.max().max() * 0.02

    for idx, row in df.reset_index().iterrows():

        row.name = row[0]
        row = row[1:].sort_values(ascending=False)[:n]

        for name in row.index:
            if name not in chat_colors.keys():
                chat_colors[name] = colors[len(chat_colors) % 10]
                handles.append(mpatches.Patch(color=chat_colors[name], label=name))

        row.plot.bar(
            width=1,
            color=[
                chat_colors[name] for name in row.index
            ],  # colordf.set_index("name").loc[row.index].color.values
            xlabel="",
            ax=axes[idx],
        )
        axes[idx].tick_params(axis="both", length=0)
        ax2 = axes[idx].twiny()
        ax2.set_xticks([0.5])
        ax2.set_xticklabels([row.name])
        if "emoji" in kwargs.keys() and emoji.emoji_count(row.name):
            offset_image([0.5, tol * 55], row.name, ax2)
            ax2.set_xticklabels([])

        ax2.tick_params(which="major", length=0)
        if label:
            for i, val in enumerate(row):
                if val:
                    axes[idx].text(
                        i,
                        val + tol,
                        int(val),
                        rotation=kwargs.get("label_rot", 90),
                        ha="center",
                    )

    if "emoji" in kwargs.keys():
        fig.suptitle(
            f"The use of Top {len(df)} used emojis across chats",
            y=legend and 0.975 or 0.92,
            fontsize=12,
            fontweight="bold",
        )
    if label:
        axes[idx].set_ylim(top=df.max().max() * 1.1)

    if legend:

        fig.legend(
            handles,
            chat_colors.keys(),
            loc=kwargs.get("loc"),
            ncol=kwargs.get("ncol", 1),
            bbox_to_anchor=kwargs.get("bbox_to_anchor"),
        )
    return axes, fig


def dt_Vs_cat(
    df,
    cat,
    metric,
    stacked,
    n=5,
    m=5,
    sort="fixed",
    aggr="count",
    cumulative=False,
    cum_aggr="sum",
    window="default",
    dt_disc=False,
    dt_hrchy="month",
    dt_groups={},
    subplots=False,
    return_data=False,
    style=False,
    label=True,
    legend=True,
    asc=False,
    **kwargs,
):

    if dt_hrchy not in freq_mapper.keys():
        raise ValueError(
            f"{dt_hrchy} not a valid datetime frequency; supported frequencies are {list(freq_mapper.keys())}"
        )
    groupping_params = {
        True: ["timestamp_ms", cat],
        False: [pd.Grouper(key="timestamp_ms", freq=freq_mapper[dt_hrchy]), cat],
    }

    # order=df.groupby(cat).agg({metric:aggr}).sort_values(metric,ascending=asc).index.values
    if dt_disc:

        df.timestamp_ms = getattr(df.timestamp_ms.dt, dt_hrchy)
        if dt_groups:
            df.timestamp_ms = df.timestamp_ms.map(
                lambda dt: dt_groups.get(dt, "Others")
            )

    df = (
        df.groupby(groupping_params[dt_disc])
        .agg({metric: aggr})
        .reset_index()
        .pivot(index="timestamp_ms", columns=cat, values=metric)
    )
    if not dt_disc:
        df = df.reindex(
            pd.date_range(df.index[0], df.index[-1], freq=freq_mapper[dt_hrchy])
        )
        df.index = df.index.date
        df.index.name = "timestamp_ms"
    df = df.fillna(0).astype("int")

    if cumulative:
        if cum_aggr not in valid_cum_aggr:
            raise ValueError(
                f"{cum_aggr} is not a valid aggregation function for cumulation; valid aggr are {valid_cum_aggr}"
            )
        if window == "default":
            window = len(df)
        df = df.rolling(window=window, min_periods=1).agg(cum_aggr)
        df = df.apply(round, args=[2])

    cum_subtitle_start = cumulative and "Cumulative" or ""
    cum_subtitle_end = (
        cumulative
        and (
            window == len(df)
            and f"(Running {cum_aggr.upper()})"
            or f"(Moving {cum_aggr.upper()} with window = {window})"
        )
        or ""
    )
    # df=df[order]
    if return_data:
        if style:
            return df.style.pipe(make_pretty)
        return df

    if sort == "fixed":
        df.columns = df.columns.map(
            lambda name: get_display(arabic_reshaper.reshape(name))
        )

    if len(df) == 1 or len(df.columns) == 1:
        ax = df.plot.bar(
            figsize=kwargs.get("figsize", (len(df) // 5 and 2 * len(df) or 10, 10)),
            legend=False,
            rot=kwargs.get("rot", 45),
            width=0.9,
            xlabel="",
        )
        if label:
            tol = df.max().values[0] * 0.02
            for idx, row in df.reset_index().iterrows():
                if row[1]:
                    ax.text(
                        idx,
                        row[1] + tol,
                        row[1],
                        ha="center",
                        rotation=kwargs.get("label_rot", 0),
                    )
            ax.set_ylim(top=tol * 55)
        # return df[:n].plot.bar(figsize=(len(df)//5 and 2*len(df) or 10,10),width=0.9,legend=legend)
        plt.title(
            f"{cum_subtitle_start} {title_freq[dt_hrchy]} Timeline of the {aggr} of {metric_display[metric]} across {cat_display[cat]} {cum_subtitle_end}",
            fontsize=12,
            fontweight="bold",
        )
        plt.ylabel(f"{aggr} of {metric_display[metric]}".upper(), fontdict={"size": 12})
        return ax
    if subplots:
        tol = df.max().max() * 0.02
        order = df.sum(0).sort_values(ascending=False).index
        df = df[order]
        fig, axes = plt.subplots(
            nrows=len(df.columns[:n]), ncols=1, sharex=True, sharey=True
        )
        plt.subplots_adjust(wspace=0, hspace=0.2)
        figsize = kwargs.get("figsize", (len(df) // 2, len(df.columns[:n]) * 2))
        for idx, sub in enumerate(df.columns[:n]):
            subset = df[sub]
            # subset.index=subset.index.date
            subset.plot.bar(
                xlabel="",
                ax=axes[idx],
                figsize=figsize,
                label=sub,
                color=colors[idx % 10],
            )
            axes[idx].legend(loc=(1.02, 0.5))
            if label:
                for i, row in subset.reset_index().iterrows():
                    if row[1]:
                        axes[idx].text(
                            i,
                            row[1] + tol,
                            row[1],
                            ha="center",
                            rotation=kwargs.get("label_rot", 0),
                        )
        if label:
            axes[idx].set_ylim(top=df.max().max() * 1.2)
        plt.xticks(rotation=kwargs.get("rot", 45))
        plt.tight_layout(pad=0, h_pad=0.2)
        fig.suptitle(
            f"{cum_subtitle_start} {title_freq[dt_hrchy]} Timeline of the {aggr} of {metric_display[metric]} across {cat_display[cat]} {cum_subtitle_end}",
            fontsize=12,
            fontweight="bold",
            y=1.02,
        )
        fig.text(
            -0.02,
            0.5,
            f"{aggr} of {metric_display[metric]}".upper(),
            rotation="vertical",
            fontsize=12,
        )
        return axes

    if stacked:
        df = df.iloc[:, :n]
        if not legend:
            df = df.sum(1)
        ax = df.plot.bar(
            figsize=kwargs.get("figsize", (len(df) // 5 and 2 * len(df) or 10, 10)),
            width=0.9,
            stacked=True,
            legend=legend,
            rot=kwargs.get("rot"),
            xlabel="",
        )
        plt.title(
            f"{cum_subtitle_start} {title_freq[dt_hrchy]} Timeline of the {aggr} of {metric_display[metric]} across {cat_display[cat]} {cum_subtitle_end}",
            fontsize=12,
            fontweight="bold",
        )
        if label:
            # tol=df.iloc[0].sum()*0.02
            # tol=df.sum(1).max()
            tol = ax.get_ylim()[1] * 0.02
            for idx, row in df.reset_index().iterrows():
                val = row[1:].sum()
                if val:
                    ax.text(
                        idx,
                        val + tol,
                        val,
                        ha="center",
                        rotation=kwargs.get("label_rot", 0),
                    )
                ax.set_ylim(top=tol * 55)
        plt.ylabel(f"{aggr} of {metric_display[metric]}".upper(), fontdict={"size": 12})
        return ax
    if sort == "dynamic":
        figsize = kwargs.get("figsize", (2 * len(df), 10))
        kwargs["figsize"] = figsize
        axes, fig = dynamic_sort(
            df,
            n=n,
            label=label,
            legend=legend,
            width=2 * len(df),
            loc="upper center",
            ncol=len(df.columns),
            bbox_to_anchor=(0.5, 0.95),
            **kwargs,
        )
        fig.suptitle(
            f"{cum_subtitle_start} {title_freq[dt_hrchy]} Timeline of the {aggr} of {metric_display[metric]} across {cat_display[cat]} {cum_subtitle_end}",
            fontsize=12,
            fontweight="bold",
            y=legend and 0.975 or 0.92,
        )
        # fig.text(0.08*(figsize[0]/15), 0.5, f"{aggr} of {metric}".upper(), va='center',ha='left', rotation='vertical')
        return axes

    figsize = kwargs.get("figsize", (len(df) // 5 and 2 * len(df) or 10, 10))
    fig, axes = plt.subplots(nrows=1, ncols=len(df), sharey=True, figsize=figsize)
    plt.subplots_adjust(wspace=0, hspace=0)
    tol = df.max().max() * 0.02
    for idx, row in df.reset_index().iterrows():
        row[1 : n + 1].plot.bar(
            width=1, color=colors[: len(df.columns[:n])], xlabel="", ax=axes[idx]
        )

        ax2 = axes[idx].twiny()
        ax2.set_xticks([0.5])
        ax2.set_xticklabels([row.timestamp_ms])
        ax2.tick_params(which="major", length=0)
        if label:
            for i, val in enumerate(row[1 : n + 1]):
                if val:
                    axes[idx].text(
                        i,
                        val + tol,
                        round(val, 2),
                        rotation=kwargs.get("label_rot", 90),
                        ha="center",
                    )
    if label:
        axes[idx].set_ylim(top=df.max().max() * 1.1)

    if legend:
        handles = []
        for idx, col in enumerate(df.columns[:n]):
            handles.append(mpatches.Patch(color=colors[idx], label=col))
        fig.legend(
            handles,
            df.columns,
            loc="upper center",
            ncol=len(df.columns),
            bbox_to_anchor=(0.5, 0.95),
        )
    # plt.xticks(rotation=kwargs.get('rot',90))

    fig.suptitle(
        f"{cum_subtitle_start} {title_freq[dt_hrchy]} Timeline of the {aggr} of {metric_display[metric]} across {cat_display[cat]} {cum_subtitle_end}",
        fontsize=12,
        fontweight="bold",
        y=legend and 0.975 or 0.93,
    )
    # fig.text(0.08*(figsize[0]/15), 0.5, f"{aggr} of {metric}".upper(),  rotation='vertical')
    return axes



import os
import shutil

cwd = os.path.dirname(__file__)


def get_sw():
    """
	Returns a list of whole stop words.
	"""

    sw_file = open(os.path.join(cwd, "sw.txt"), "r", encoding="utf8")
    data = sw_file.read()

    sw = data.split("\n")
    sw_file.close()
    return sw


def add_words(words):
    """
	Used to add stop words to the sw list.

	Parameters
	----------
	words : list
		the new words to append to your current list.
	"""
    if not isinstance(words, list):
        raise TypeError(f"expected a list, got {type(words)}")
    backup = os.path.join(cwd, "sw_backup.txt")
    if not os.path.isfile(backup):
        shutil.copy(os.path.join(cwd, "sw.txt"), backup)

    sw = set(get_sw())
    with open(os.path.join(cwd, "sw.txt"), "a", encoding="utf8") as myfile:
        for word in words:
            if word not in sw:
                myfile.write("\n" + word)


def reset_sw():
    """
	Reset your sw list to default.	
	"""
    default_sw = open(os.path.join(cwd, "sw_backup.txt"), "r", encoding="utf8")
    with open(os.path.join(cwd, "sw.txt"), "w", encoding="utf8") as myfile:
        myfile.write(default_sw.read())
    default_sw.close()


def clear_sw(clear_backup=False):
    """
	Clear the whole list of stop words.

	Parameters
	----------
	clear_backup : bool, default False
		Whether to clear the backed up list or not.
		Please note that if you cleared the backed up list,
		you won't be able to reset to the default sw again.
	"""
    backup = os.path.join(cwd, "sw_backup.txt")
    if not os.path.isfile(backup) and not clear_backup:
        shutil.copy(os.path.join(cwd, "sw.txt"), backup)
    elif clear_backup and os.path.isfile(backup):

        with open(backup, "w", encoding="utf8") as myfile:
            myfile.write("")

    with open(os.path.join(cwd, "sw.txt"), "w", encoding="utf8") as myfile:
        myfile.write("")



from .common_plots import plot, multicat_plot, chat_plot
from .plot_helper import top_timely, metric_correct
from .media import media_plot
from pandas.util._decorators import doc


class share_plot(multicat_plot):
    @doc(media_plot.bar, cat="domain", alias="link domain", metric="links")
    def bar(
        self,
        by_domain=False,
        n_domain=5,
        include_domain=[],
        exclude_domain=[],
        domain_groups={},
        domain_others=False,
        per_chat=False,
        over_time=False,
        n=5,
        aggr="count",
        cumulative=False,
        cum_aggr="sum",
        window="default",
        include=[],
        exclude=[],
        chat_groups={},
        others=False,
        dt_disc=False,
        dt_hrchy="month",
        dt_range="last year",
        dt_groups={},
        subplots=False,
        stacked="default",
        return_data=False,
        style=False,
        label=True,
        legend=True,
        asc=False,
        show_sender=True,
        sort="fixed",
        **kwargs
    ):
        cat = "domain"
        return super(share_plot, self).bar(
            by_domain,
            n_domain,
            include_domain,
            exclude_domain,
            domain_groups,
            domain_others,
            per_chat,
            over_time,
            n,
            aggr,
            cumulative,
            cum_aggr,
            window,
            include,
            exclude,
            chat_groups,
            others,
            dt_disc,
            dt_hrchy,
            dt_range,
            dt_groups,
            subplots,
            stacked,
            return_data,
            style,
            label,
            legend,
            asc,
            show_sender,
            sort,
            cat,
            **kwargs
        )

    @doc(media_plot.line, cat="domain", alias="link domain", metric="links")
    def line(
        self,
        by_domain=False,
        per_chat=False,
        n=5,
        aggr="count",
        cumulative=False,
        cum_aggr="sum",
        window="default",
        include=[],
        exclude=[],
        groups={},
        asc=False,
        show_sender=False,
        others=False,
        dt_disc=False,
        dt_hrchy="month",
        dt_range="last year",
        subplots=False,
        return_data=False,
        style=False,
        legend=True,
        area=False,
        **kwargs
    ):
        cat = "domain"
        return super(share_plot, self).line(
            by_domain,
            per_chat,
            n,
            aggr,
            cumulative,
            cum_aggr,
            window,
            include,
            exclude,
            groups,
            asc,
            show_sender,
            others,
            dt_disc,
            dt_hrchy,
            dt_range,
            subplots,
            return_data,
            style,
            legend,
            area,
            cat,
            **kwargs
        )

    @doc(media_plot.pie, cat="domain", alias="link domain", metric="links")
    def pie(
        self,
        by_domain=False,
        n=5,
        include=[],
        exclude=[],
        groups={},
        others=True,
        return_data=False,
        aggr="count",
        label=True,
        **kwargs
    ):
        cat = "domain"
        return super(share_plot, self).pie(
            by_domain,
            n,
            include,
            exclude,
            groups,
            others,
            return_data,
            aggr,
            label,
            cat,
            **kwargs
        )

    def dot(
        self,
        n_domain=5,
        include_domain=[],
        exclude_domain=[],
        domain_groups={},
        domain_others=False,
        x="month",
        y="day",
        disc=(0, 1),
        aggr="count",
        showcmap=True,
        size=200,
        clip=None,
        include=[],
        n=5,
        exclude=[],
        chat_groups={},
        legend=False,
        others=False,
        return_data=False,
        style=False,
        dt_range="last year",
        **kwargs
    ):

        cat = "domain"
        return super(share_plot, self).dot(
            n_domain,
            include_domain,
            exclude_domain,
            domain_groups,
            domain_others,
            x,
            y,
            disc,
            aggr,
            showcmap,
            size,
            clip,
            include,
            n,
            exclude,
            chat_groups,
            legend,
            others,
            return_data,
            style,
            dt_range,
            cat,
            **kwargs
        )

    @doc(media_plot.gantt_chart, cat="domain", alias="link domain", metric="links")
    def gantt_chart(
        self,
        by_domain=False,
        msg_freq="day",
        ylabel_freq="month",
        dt_range="last year",
        n=5,
        include=[],
        exclude=[],
        groups={},
        others=False,
        return_data=False,
        style=False,
        **kwargs
    ):
        return super(share_plot, self).gantt_chart(
            by_domain,
            msg_freq,
            ylabel_freq,
            dt_range,
            n,
            include,
            exclude,
            groups,
            others,
            return_data,
            style,
            cat="domain",
            **kwargs
        )


class chat_share_plot(plot):
    def bar(
        self,
        by_domain=False,
        over_time=False,
        n_domain=5,
        aggr="count",
        cumulative=False,
        cum_aggr="sum",
        window="default",
        include_domain=[],
        exclude_domain=[],
        domain_groups={},
        others=False,
        dt_disc=False,
        dt_hrchy="month",
        dt_range="last year",
        dt_groups={},
        subplots=False,
        stacked="default",
        return_data=False,
        style=False,
        label=True,
        legend=True,
        asc=False,
        show_sender=True,
        sort="fixed",
        **kwargs
    ):
        """
        Make vertical bar plot/s of chat links data using link domain as its categorical variable.

        See stats.Links.plot.bar
        """
        cat = by_domain and "domain" or "sender_name"
        per_chat = by_domain or False
        include = by_domain and include_domain or []
        exclude = by_domain and exclude_domain or []
        n = by_domain and n_domain or 0
        chat_groups = by_domain and domain_groups or {}

        return super(chat_share_plot, self).bar(
            per_chat,
            over_time,
            n,
            aggr,
            cumulative,
            cum_aggr,
            window,
            include,
            exclude,
            chat_groups,
            others,
            dt_disc,
            dt_hrchy,
            dt_range,
            dt_groups,
            subplots,
            stacked,
            return_data,
            style,
            label,
            legend,
            asc,
            show_sender,
            sort,
            cat=cat,
            **kwargs
        )

    def line(
        self,
        by_domain=False,
        n_domain=5,
        aggr="count",
        cumulative=False,
        cum_aggr="sum",
        window="default",
        include_domain=[],
        exclude_domain=[],
        domain_groups={},
        asc=False,
        show_sender=False,
        others=False,
        dt_disc=False,
        dt_hrchy="month",
        dt_range="last year",
        subplots=False,
        return_data=False,
        style=False,
        legend=True,
        area=False,
        **kwargs
    ):
        """
        Make line chart/s of chat links data using link domain as its categorical variable.

        See stats.Links.plot.line
        """
        cat = by_domain and "domain" or "sender_name"
        per_chat = by_domain or False
        include = by_domain and include_domain or []
        exclude = by_domain and exclude_domain or []
        n = by_domain and n_domain or 0
        chat_groups = by_domain and domain_groups or {}
        return super(chat_share_plot, self).line(
            per_chat,
            n,
            aggr,
            cumulative,
            cum_aggr,
            window,
            include,
            exclude,
            chat_groups,
            asc,
            show_sender,
            others,
            dt_disc,
            dt_hrchy,
            dt_range,
            subplots,
            return_data,
            style,
            legend,
            area,
            cat=cat,
            **kwargs
        )

    def pie(
        self,
        by_domain=False,
        n_domain=5,
        include_domain=[],
        exclude_domain=[],
        domain_groups={},
        others=True,
        return_data=False,
        aggr="count",
        label=True,
        **kwargs
    ):
        """
        Make pie chart of chat links data using link domain as its categorical variable.

        See stats.Links.plot.pie
        """
        cat = by_domain and "domain" or "sender_name"
        per_chat = by_domain or False
        include = by_domain and include_domain or []
        exclude = by_domain and exclude_domain or []
        n = by_domain and n_domain or 0
        chat_groups = by_domain and domain_groups or {}
        return super(chat_share_plot, self).pie(
            n,
            include,
            exclude,
            chat_groups,
            others,
            return_data,
            aggr,
            label,
            cat=cat,
            **kwargs
        )

    def dot(
        self,
        x="month",
        y="day",
        disc=(0, 1),
        aggr="count",
        showcmap=True,
        size=200,
        clip=None,
        include_domain=[],
        n_domain=5,
        exclude_domain=[],
        domain_groups={},
        legend=False,
        others=False,
        return_data=False,
        style=False,
        dt_range="last year",
        **kwargs
    ):
        """
        Make dot plot of chat links data using link domain as its categorical variable.

        See stats.Links.plot.dot
        """
        by_domain = x == "domain"
        cat = by_domain and "domain" or "sender_name"
        include = by_domain and include_domain or []
        exclude = by_domain and exclude_domain or []
        n = by_domain and n_domain or 0
        chat_groups = by_domain and domain_groups or {}
        return super(chat_share_plot, self).dot(
            x,
            y,
            disc,
            aggr,
            showcmap,
            size,
            clip,
            include,
            n,
            exclude,
            chat_groups,
            legend,
            others,
            return_data,
            style,
            dt_range,
            cat=cat,
            **kwargs
        )

    def gantt_chart(
        self,
        by_ext=False,
        msg_freq="day",
        ylabel_freq="month",
        dt_range="last year",
        include=[],
        n=5,
        exclude=[],
        groups={},
        others=False,
        return_data=False,
        style=False,
        **kwargs
    ):
        """
        Make gantt chart of chat links data using link domain as its categorical variable.

        See stats.Links.plot.gantt_chart
        """
        return super(chat_share_plot, self).gantt_chart(
            msg_freq,
            ylabel_freq,
            dt_range,
            n,
            include,
            exclude,
            groups,
            others,
            return_data,
            style,
            cat="domain",
            **kwargs
        )

    @doc(chat_plot.ranking)
    def ranking(
        self,
        n=15,
        dt_hrchy="day",
        cat="sender_name",
        aggr="count",
        dt_range=None,
        show_name=False,
        legend=True,
        rank=False,
        **kwargs
    ):
        metric = metric_correct(self._metric, aggr)
        return top_timely(
            self.data,
            metric,
            aggr,
            cat,
            n,
            dt_hrchy,
            dt_range,
            show_name,
            legend,
            rank,
            **kwargs
        )



import emoji
import arabic_reshaper
from bidi.algorithm import get_display
from wordcloud import WordCloud
import numpy as np
import matplotlib.pyplot as plt
from .special_plots import BubbleChart
from .plot_helper import get_meta_data, cat_filter
from .SW import get_sw

stops = get_sw()


def message_filter(data, n_words=2, emoji_strip=False, subset=None):
    data = data.copy()
    if subset:
        meta_data = get_meta_data()
        if subset == "me":
            data = data[data.sender_name == meta_data["user"]]
        elif subset == "others":
            data = data[~(data.sender_name == meta_data["user"])]
    if emoji_strip:
        if len(data.shape) < 2:
            data = data.dropna()
            # data=data.map(lambda text: emoji.replace_emoji(text, "")).str.strip().replace('',np.nan).dropna()
        else:

            # data['content']=data['content_striped']
            # data=data.dropna(subset='content')
            data = data.dropna(subset="content_striped")
            data["content"] = data["content_striped"]
            # data.content=data.content.map(lambda text: emoji.replace_emoji(text, "")).str.strip().replace('',np.nan)

    if n_words:
        if len(data.shape) < 2:
            data = data[data.str.split().apply(len) <= n_words]
        else:
            data = data.content[data.content.str.split().apply(len) <= n_words]
    return data


def trim_counter_sw(counter):
    for stop in stops:
        if stop in counter.keys():
            del counter[stop]
    return counter


def word(data, width=800, height=400):
    data.index = data.index.map(lambda name: get_display(arabic_reshaper.reshape(name)))
    frequencies = dict(data)
    return WordCloud(
        font_path="arial", width=width, height=height
    ).generate_from_frequencies(frequencies)


def word_cloud_single(data, title, width=1600, height=800):
    if not len(data):
        raise ValueError("No data to plot")
    wordcloud = word(data, width=width, height=height)
    plt.figure(figsize=(15, 10), facecolor="k")
    plt.title(title, fontsize=30, y=1, color="white")
    fig = plt.imshow(wordcloud)
    plt.axis("off")
    plt.tight_layout(pad=0)

    return fig


def word_cloud_sender(data_lst, titles, suptitle, width=800, height=280):
    fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(15, 10))
    fig.suptitle(suptitle, y=1, fontsize=30)
    plt.subplots_adjust(wspace=0, hspace=0)
    plt.tight_layout(pad=0, h_pad=0)
    for idx, data in enumerate(data_lst):
        if not len(data):
            continue
        ax = axes[idx]
        wordcloud = word(data, width=width, height=height)
        ax.imshow(wordcloud)
        ax.set_title(
            titles[idx], x=-0.02, y=0.2, rotation=90, fontdict={"fontsize": 20}
        )
        ax.axis("off")
    return axes


def word_cloud_time(data_lst, titles, suptitle, width=800, height=300):

    rows = len(data_lst) % 3 and 1 + len(data_lst) // 3 or int(len(data_lst) / 3)
    fig, axes = plt.subplots(ncols=3, nrows=rows, figsize=(30, rows * 4))
    fig.suptitle(suptitle, y=1, fontsize=30)
    plt.subplots_adjust(wspace=0, hspace=0)
    try:
        titles = list(
            map(lambda title: get_display(arabic_reshaper.reshape(title)), titles)
        )
    except:
        pass
    axes = axes.flatten()
    for idx, data in enumerate(data_lst):
        if not len(data):
            continue
        ax = axes[idx]
        wordcloud = word(data, width, height)
        ax.set_title(
            titles[idx],
            x=-0.02,
            y="chat" in suptitle and 0.05 or 0.4,
            rotation=90,
            fontdict={"fontsize": "chat" in suptitle and 15 or 20},
        )
        ax.imshow(wordcloud)

    for ax in axes:
        ax.axis("off")
    plt.tight_layout(
        pad=("chat" in suptitle and rows < 3) and 1 or 2,
        h_pad="chat" in suptitle and 1 or 1 / rows,
    )

    return axes


def bubble_data(data):
    data = data.reset_index()
    data.columns = ["word", "count"]
    data = data.sample(len(data), replace=False).reset_index()
    data.word = data.word.map(lambda name: get_display(arabic_reshaper.reshape(name)))
    return data


def bubble_single(data, title):
    data = bubble_data(data)
    if len(data) < 2:
        raise ValueError(
            f"Bubble Chart need at least 2 data points, got only {len(data)}"
        )
    bubble_chart = BubbleChart(area=data["count"], bubble_spacing=2)
    bubble_chart.collapse()
    fig, ax = plt.subplots(subplot_kw=dict(aspect="equal"))
    ax.set_title(title, fontsize=30)
    fig.set_size_inches(20, 20, forward=False)
    bubble_chart.plot(
        ax,
        data["word"],
        [plt.colormaps["tab20"](i) for i in np.random.choice(range(20), len(data))],
        fontsize=14,
        show_counts=True,
    )
    ax.axis("off")
    ax.relim()
    ax.autoscale_view()

    return ax


def bubble_sender(data_lst, titles, suptitle):

    print(data_lst)
    fig, axes = plt.subplots(ncols=2, nrows=1, figsize=(25, 12.5))
    fig.suptitle(suptitle, fontsize=30, y=1.04)
    plt.subplots_adjust(wspace=0, hspace=0)
    plt.tight_layout(pad=0)
    for idx, data in enumerate(data_lst):
        if len(data) < 2:
            continue
        data = bubble_data(data)
        bubble_chart = BubbleChart(area=data["count"], bubble_spacing=1)
        bubble_chart.collapse()
        ax = bubble_chart.plot(
            axes[idx],
            data["word"],
            [plt.colormaps["tab20"](i) for i in np.random.choice(range(20), len(data))],
            fontsize=14,
            show_counts=True,
        )
        ax.axis("off")
        ax.relim()
        ax.autoscale_view()
        ax.set_title(titles[idx].upper(), fontdict={"fontsize": 20})

    return axes


def bubble_time(data_lst, titles, suptitle):

    rows = len(data_lst) % 3 and 1 + len(data_lst) // 3 or int(len(data_lst) / 3)
    fig, axes = plt.subplots(ncols=3, nrows=rows, figsize=(30, rows * 10))
    axes = axes.flatten()
    fig.suptitle(suptitle, y=1 + 0.1 / rows, fontsize=30)
    plt.subplots_adjust(wspace=0, hspace=0)
    plt.tight_layout(pad=0)
    try:
        titles = list(
            map(lambda title: get_display(arabic_reshaper.reshape(title)), titles)
        )
    except:
        pass
    for idx, data in enumerate(data_lst):
        if len(data) < 2:
            continue
        data = bubble_data(data)

        bubble_chart = BubbleChart(area=data["count"], bubble_spacing=0.1)
        bubble_chart.collapse()
        ax = bubble_chart.plot(
            axes[idx],
            data["word"],
            [plt.colormaps["tab20"](i) for i in np.random.choice(range(20), len(data))],
            fontsize=16,
            show_counts=True,
        )
        ax.set_title(titles[idx], fontdict={"fontsize": 20})
        # ax.relim()
        ax.autoscale_view()
    for ax in axes:
        ax.axis("off")

    return axes


charts = {
    "single": {"bubble": bubble_single, "word_cloud": word_cloud_single},
    "sender": {"bubble": bubble_sender, "word_cloud": word_cloud_sender},
    "time": {"bubble": bubble_time, "word_cloud": word_cloud_time},
}



from setuptools import find_packages, setup
from codecs import open
from os import path

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


with open(path.join(HERE, "requirements.txt"), "r") as f:
    requirements = f.read().splitlines()

setup(
    name='socialysis',
    packages=find_packages(include=['socialysis']),
    include_package_data=True,
    version='0.0.0',
    url="https://github.com/abdulrahmankhayal/socialysis",
    description='Tool for analyzing and extracting insights from Facebook Messenger conversations',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Abdul-Rahman Khayyal',
    license='MIT',
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    install_requires=requirements,
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['./stickers_used/*','./sw.txt','SAMPLE_DATA.pkl','sample_meta_data.json']
    }
)
